/*
 * Created on Oct 21, 2004
 */
package net.kildenpedersen.reflect;

import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.lang.reflect.AccessibleObject;
import java.lang.reflect.Array;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Member;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;

/**
 * @author Nils Kilden-Pedersen
 */
public final class Reflector {

	private static final String DOT = ".";
    
    private static Set getAccessibleMethods(final Class type, final String name, int parmCount, boolean staticMethods) {
        Method[] methods = type.getDeclaredMethods();
        HashSet methodSet = new HashSet(methods.length);
        for (int i = 0; i < methods.length; i++) {
            if (methods[i].getName().equals(name) 
                    && methods[i].getParameterTypes().length == parmCount 
                    && staticMethods == Modifier.isStatic(methods[i].getModifiers())) {
                setAccessible(methods[i]);
                methodSet.add(methods[i]);
            }
        }
        return methodSet;
    }
    /**
     * @param target
     * @param propertyName
     * @param methodType "set" or "get"
     * @param propertyType
     * @return method
     */
    private static Method getAccessiblePropertyMethod(
            final Class target, 
            String propertyName, 
            String methodType, 
            Class propertyType) {
        
        Method method = null;
        String methodName = methodType + propertyName.substring(0, 1).toUpperCase() + 
    		(propertyName.length() > 1 ? propertyName.substring(1) : "");
        Method[] methods = target.getDeclaredMethods();
        for (int i = 0; i < methods.length; i++) {
			if (methods[i].getParameterTypes().length == 1 
					&& methods[i].getName().equals(methodName)) {
			    method = methods[i];
				if (propertyType != null 
				        && propertyType.equals(method.getParameterTypes()[0])) {
				    break;
				}
			}
        }
        setAccessible(method);
        return method;
    }
    
	/**
	 * Get named node value.
	 * @param nodeObject
	 * @param nodeName
	 * @return node value
	 */
	private static Object getNodeValue(Object nodeObject, final String nodeName) {
	
	    String parmName = nodeName;
	    if (parmName.endsWith("]")) {
	        int pos = parmName.lastIndexOf('[');
	        if (pos + 2 == parmName.length()) {
	        	throw new IllegalArgumentException("Index empty: " + nodeName);
	        }
	        int idx = Integer.parseInt(
	                parmName.substring(pos+1, parmName.length()-1)
	                );
	        parmName = parmName.substring(0, pos);
	        nodeObject = getNodeValue(nodeObject, parmName);
	        try {
	        	return getIndexObject(nodeObject, idx);
	        } catch (IllegalArgumentException e) {
		        String msg = parmName + "[] is not array or java.util.List";
		        throw new IllegalArgumentException(msg);
	        }
	    }
	    
		// Check if it's a Map
		if (nodeObject instanceof Map) {
			Map parameterMap = (Map) nodeObject;
			// Must check for key. Value can be null. 
			if (!parameterMap.containsKey(parmName)) {
			    throw newException(parmName, nodeName);    		}
			return parameterMap.get(parmName);
		}
	
		Class parmClass = nodeObject.getClass();
		Method readMethod;
	
		if (parmName.endsWith("()")) {
			parmName = parmName.substring(0, parmName.length()-2);
		    readMethod = Reflector.getAccessibleMethod(parmClass, parmName, null);
		    if (readMethod == null) {
		        String msg = "No " + parmName + " method found on " + parmClass.toString();
		        throw new IllegalArgumentException(msg);
		    }
		} else {
		    readMethod = Reflector.getAccessibleReadProperty(parmClass, parmName);
		}
	
		if (readMethod != null) {
			try {
				return readMethod.invoke(nodeObject, null);
			} catch (Exception e) {
				throw new RuntimeException(e);
			}
		}
	
		// If not method, try field
		Field field = Reflector.getAccessibleField(parmClass, parmName);
		if (field != null) {
			try {
				return field.get(nodeObject);
			} catch (Exception e) {
				throw new RuntimeException(e);
			}
		}
		
		// If all else fails
		throw newException(parmName, nodeName);
	}

    private static PropertyDescriptor getProperty(Class target, String name) {
        PropertyDescriptor[] properties;
        try {
            properties = Introspector.getBeanInfo(target, Object.class).getPropertyDescriptors();
        } catch (IntrospectionException e) {
            throw new RuntimeException(e);
        }
        for (int i = 0; i < properties.length; i++) {
            if (name.equals(properties[i].getName())) {
                return properties[i];
            }
        }
        return null;
    }
    private static IllegalArgumentException newException(String node, String tree) {
        String ending = (node.equals(tree) ? "'" : "' in '" + tree + "'");
        return new IllegalArgumentException("Cannot find '" + node + ending);
    }
    
    private static void setAccessible(AccessibleObject object) {
        if (object != null && !object.isAccessible()) {
            object.setAccessible(true);
        }
    }
    
    /**
     * @param type
     * @param parmTypes
     * @return accessible constructor
     */
    public static Constructor getAccessibleConstructor(final Class type, Class[] parmTypes) {
        Constructor constructor;
        try {
            constructor = type.getDeclaredConstructor(parmTypes);
        } catch (NoSuchMethodException e) {
            return null;
        }
        setAccessible(constructor);
        return constructor;
    }
    
    /**
     * Get a Set of accessible constructors with the given parameter count.
     * @param type
     * @param parmCount
     * @return constructor array
     */
    public static Set getAccessibleConstructors(final Class type, int parmCount) {
        Constructor[] constructors = type.getDeclaredConstructors();
        HashSet constructorSet = new HashSet(constructors.length);
        for (int i = 0; i < constructors.length; i++) {
            if (constructors[i].getParameterTypes().length == parmCount) {
                setAccessible(constructors[i]);
                constructorSet.add(constructors[i]);
            }
        }
        return constructorSet;
    }
    
    /**
     * @param type
     * @param name
     * @return field or null.
     */
    public static Field getAccessibleField(final Class type, final String name) {
        Field field;
        try {
            field = type.getDeclaredField(name);
        } catch (NoSuchFieldException e) {
            return null;
        }
        setAccessible(field);
        return field;
    }

    /**
     * Return set of accessible member methods with the given name and parameter count.
     * @param type class type
     * @param name method name
     * @param parmCount parameter count
     * @return Set of {@linkplain Method}s
     */
    public static Set getAccessibleMemberMethods(final Class type, final String name, final int parmCount) {
        return getAccessibleMethods(type, name, parmCount, false);
    }
    
    /**
     * @param type
     * @param name
     * @param parmTypes
     * @return accessible method
     */
    public static Method getAccessibleMethod(final Class type, final String name, final Class[] parmTypes) {
        Method method;
        try {
            method = type.getDeclaredMethod(name, parmTypes);
        } catch (NoSuchMethodException e) {
            return null;
        }
        setAccessible(method);
        return method;
    }
    
    /**
     * @param target
     * @param name property name
     * @return accessible read property
     */
    public static Method getAccessibleReadProperty(final Class target, String name) {
        Method readMethod = null;
        Class propertyType = null;
        PropertyDescriptor property = getProperty(target, name);
        if (property != null) {
            readMethod = property.getReadMethod();
            if (readMethod != null) {
                return readMethod;
            }
            propertyType = property.getPropertyType();
        }
        
        readMethod = getAccessiblePropertyMethod(target, name, "get", propertyType);
        if (readMethod == null) {
            readMethod = getAccessiblePropertyMethod(target, name, "is", propertyType);
        }
        return readMethod;
    }
    
    /**
     * Return set of accessible static methods with the given name and parameter count.
     * @param type class type
     * @param name method name
     * @param parmCount parameter count
     * @return Set of {@linkplain Method}s

     */
    public static Set getAccessibleStaticMethods(final Class type, final String name, int parmCount) {
        return getAccessibleMethods(type, name, parmCount, true);
    }

    /**
     * @param target target
     * @param name property name
     * @return accessible read property
     */
    public static Method getAccessibleWriteProperty(final Class target, String name) {
        Method writeMethod = null;
        Class propertyType = null;
        PropertyDescriptor property = getProperty(target, name);
        if (property != null) {
            writeMethod = property.getWriteMethod();
            if (writeMethod != null) {
                return writeMethod;
            }
            propertyType = property.getPropertyType();
        }
        
        writeMethod = getAccessiblePropertyMethod(target, name, "set", propertyType);
        return writeMethod;
    }
    /**
     * Get class name, corrected for arrays
     * @param type
     * @return class name string
     */
    public static String getClassName(Class type) {
        return type.isArray() ? 
                type.getComponentType().getName()+"[]" : 
                    type.getName();
    }
    
    /**
     * Return object at the given index.
     * @param collection Array or List of objects
     * @param index Index value to return
     * @return indexed value
     * @throws IllegalArgumentException if object is not array or collection
     */
    public static Object getIndexObject(Object collection, int index) {
    	if (collection.getClass().isArray()) {
    		return Array.get(collection, index);
    	}
    	if (collection instanceof List) {
    		return ((List) collection).get(index);
    	}
    	throw new IllegalArgumentException("Object not array or List.");
    }
    
    /**
     * Get leaf value on a dot separated tree.
     * @param rootObject
     * @param nodeTree
     * @return object value
     */
    public static Object getLeafValue(final Object rootObject, final String nodeTree) {
    
    	if (rootObject == null) {
    		return null;
    	}
    	
    	if (nodeTree.indexOf(DOT) < 0) {
    		return getNodeValue(rootObject, nodeTree);
    	}
    	
    	Object node = rootObject;
		StringTokenizer tokens = new StringTokenizer(nodeTree, DOT);
		while (tokens.hasMoreTokens() && node != null) {
			node = Reflector.getNodeValue(node, tokens.nextToken());
		}
		return node;

    }

    /**
     * Get the parameter types for the member.
     * @param member
     * @return array of parameter types
     */
    public static Class[] getParameterTypes(Member member) {
    	/*
    	 * Damn Sun for not having a supertype or 
    	 * interface for both constructors and methods 
    	 * with a getParameterTypes() method.
    	 */
    	Class memberType = member.getClass();
    	if (memberType == Constructor.class) {
    		return ((Constructor) member).getParameterTypes();
    	} else if (memberType == Method.class) {
    		return ((Method) member).getParameterTypes();
    	}
    	throw new IllegalArgumentException("Member type is neither constructor nor method. This is not your fault, but the idiot who programmed this.");
    }

    /**
     * Get size of array or collection.
     * @param collection Array or collection of objects
     * @return size of collection
     */
    public static int getSize(Object collection) {
    	if (collection.getClass().isArray()) {
    		return Array.getLength(collection);
    	}
    	if (collection instanceof Collection) {
    		return ((Collection) collection).size();
    	}
    	throw new IllegalArgumentException("Object not array or Collection.");
    }

	/**
	 * Set leaf value on a dot separated tree.
	 * @param rootObject Root object
	 * @param nodeTree Node tree
	 * @param value Value to set
	 */
	public static void setLeafValue(
			final Object rootObject, 
			final String nodeTree, 
			final Object value) {
	
		Object settableObject;
		String settableName;
		int lastDot = nodeTree.lastIndexOf(DOT);
		if (lastDot > 0) {
			settableObject = getLeafValue(rootObject, nodeTree.substring(0, lastDot));
			settableName = nodeTree.substring(lastDot+1);
		} else {
			settableObject = rootObject;
			settableName = nodeTree;
		}

	    if (settableName.endsWith("]")) {
	        int pos = settableName.lastIndexOf('[');
	        int idx = Integer.parseInt(
	        		settableName.substring(pos+1, settableName.length()-1)
	                );
	        settableName = settableName.substring(0, pos);
	        settableObject = getNodeValue(settableObject, settableName);
	        if (settableObject.getClass().isArray()) {
	        	Array.set(settableObject, idx, value);
	            return;
	        }
	        if (settableObject instanceof List) {
	            List list = (List) settableObject;
	            list.set(idx, value);
	            return;
	        } 
	        String msg = settableName + "[] is not array or java.util.List";
	        throw new IllegalArgumentException(msg);
	    }
	    
		// Check if it's a Map
		if (settableObject instanceof Map) {
			Map map = (Map) settableObject;
			map.put(settableName, value);
			return;
		}
	
		Class settableClass = settableObject.getClass();
		Method writeMethod;
	
		if (settableName.endsWith("(*)")) {
			settableName = settableName.substring(0, settableName.length()-3);
		    writeMethod = Reflector.getAccessibleMethod(settableClass, settableName, new Class[] {value.getClass()});
		    if (writeMethod == null) {
		        String msg = "No '" + settableName + "' method found on " + settableClass.toString();
		        throw new IllegalArgumentException(msg);
		    }
		} else {
		    writeMethod = Reflector.getAccessibleWriteProperty(settableClass, settableName);
		}
	
		if (writeMethod != null) {
			try {
				writeMethod.invoke(settableObject, new Object[] {value});
				return;
			} catch (Exception e) {
				throw new RuntimeException(e);
			}
		}
	
		// If not method, try field
		Field field = Reflector.getAccessibleField(settableClass, settableName);
		if (field != null) {
			try {
				field.set(settableObject, value);
				return;
			} catch (Exception e) {
				String msg;
				Class fieldType = field.getType();
				if (fieldType.isPrimitive() && value == null) {
					msg = "Cannot set primitive field '" + field.getName() + "' with null value.";
				} else if (value != null && !fieldType.isAssignableFrom(value.getClass())) {
					msg = "Cannot assign type " + value.getClass().getName() + 
					" to field '" + field.getName() + "' of type " + fieldType.getName();
				} else {
					msg = e.getMessage();
				}
				throw new RuntimeException(msg, e);
			}
		}
		
		// If all else fails
		throw newException(settableName, nodeTree);
	}
    
    /**
     * Private constructor.
     */
    private Reflector() {
        return;
    }
}
