package net.kildenpedersen.xml;

import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/*
 * Created on Mar 20, 2004
 *  
 */

/**
 * Builds a DOM document. Performs XInclude and schema full validation if Xerces
 * parser is present.
 * 
 * @author Nils Kilden-Pedersen
 */
public class DOMBuilder {

    private class StrictErrorHandler implements org.xml.sax.ErrorHandler {

        /**
         * @see org.xml.sax.ErrorHandler#error(SAXParseException)
         */
        public void error(SAXParseException e) throws SAXException {
            throw e;
        }

        /**
         * @see org.xml.sax.ErrorHandler#fatalError(SAXParseException)
         */
        public void fatalError(SAXParseException e) throws SAXException {
            throw e;
        }

        /**
         * @see org.xml.sax.ErrorHandler#warning(SAXParseException)
         */
        public void warning(SAXParseException e) throws SAXException {
            throw e;
        }
    }

    //private static final String XERCES_CONFIG_CLASS_KEY = "org.apache.xerces.xni.parser.XMLParserConfiguration";

    private static final String XERCES_DOM_IMPL = "org.apache.xerces.dom.DOMImplementationImpl";
/*
    private static final String XERCES_XINCLUDE_CONFIG_CLASS;
    static {
        byte[] xercesVersion = getXercesVersion();
        if (xercesVersion != null && xercesVersion[0] >= 2 && xercesVersion[1] >= 6) {
            XERCES_XINCLUDE_CONFIG_CLASS =
                "org.apache.xerces.parsers.XIncludeParserConfiguration";
        } else {
            XERCES_XINCLUDE_CONFIG_CLASS = null;
        }
    }
*/
    private static String getDOMImplementation(DocumentBuilderFactory factory) {
        try {
            return factory.newDocumentBuilder().getDOMImplementation()
                    .getClass().getName();
        } catch (ParserConfigurationException e) {
            return "";
        }
    }
/*
    private static byte[] getXercesVersion() {
        String versionString;
        try {
            versionString = (String) 
            	Class.forName("org.apache.xerces.impl.Version").
            		getMethod("getVersion", null).
            			invoke(null, null);
        } catch (Exception e) {
            return null;
        }
        int spacePos = versionString.indexOf(' ');
        if (spacePos < 0) {
            return null;
        }
        versionString = versionString.substring(spacePos);
        byte[] version = new byte[3];
        StringTokenizer tokens = new StringTokenizer(versionString, " .");
        byte pos = 0;
        while (tokens.hasMoreTokens()) {
            String token = tokens.nextToken();
            try {
                version[pos++] = Byte.parseByte(token);
            } catch (NumberFormatException e) {
                return null;
            }
        }
        return version;

    }
*/
    
    /**
     * Return the first Element
     * 
     * @param node
     * @return The first child element or null if no children exist.
     */
    public static Element getFirstChildElement(Node node) {
        Node childNode = node.getFirstChild();
        do {
            if (childNode.getNodeType() == Node.ELEMENT_NODE) {
                return (Element) childNode;
            }
        } while ((childNode = childNode.getNextSibling()) != null);
        return null;
    }

    private DocumentBuilder builder = null;

    /**
     * Constructor.
     *  
     */
    public DOMBuilder() {
        this(null);
    }

    /**
     * Constructor.
     * 
     * @param schemaURL
     */
    public DOMBuilder(URL schemaURL) {
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        factory.setNamespaceAware(true);
        String domImpl = getDOMImplementation(factory);
        if (schemaURL != null && XERCES_DOM_IMPL.equals(domImpl)) {
            factory.setAttribute(
                    "http://apache.org/xml/features/validation/dynamic",
                    Boolean.TRUE);
            factory.setAttribute(
                    "http://apache.org/xml/features/validation/schema",
                    Boolean.TRUE);
            factory.setAttribute(
                    "http://apache.org/xml/features/validation/schema-full-checking",
                    Boolean.TRUE);
            factory.setAttribute(
                    "http://apache.org/xml/properties/schema/external-noNamespaceSchemaLocation",
                    schemaURL.toExternalForm());
        }

/*
        // Save any existing config definition
        String savedConfigClass = System.getProperty(XERCES_CONFIG_CLASS_KEY);
        if (XERCES_XINCLUDE_CONFIG_CLASS != null) {
            System.setProperty(XERCES_CONFIG_CLASS_KEY,
                    XERCES_XINCLUDE_CONFIG_CLASS);
        }
*/

        try {
            this.builder = factory.newDocumentBuilder();
        } catch (ParserConfigurationException e) {
            throw new RuntimeException(e);
        }
/*
        // Reinstate previous config definition, if it existed, otherwise clear
        if (savedConfigClass != null) {
            System.setProperty(XERCES_CONFIG_CLASS_KEY, savedConfigClass);
        } else if(XERCES_XINCLUDE_CONFIG_CLASS != null) {
            System.getProperties().remove(XERCES_CONFIG_CLASS_KEY);
        }
*/
        this.builder.setErrorHandler(new StrictErrorHandler());
    }

    /**
     * Parse input stream.
     * 
     * @param xmlDocument
     * @return DOM document
     */
    public Document parse(InputStream xmlDocument) {
        if (xmlDocument == null) {
            throw new IllegalArgumentException("XML document is null.");
        }
        try {
            return this.builder.parse(xmlDocument);
        } catch (SAXException e) {
            throw new RuntimeException(e);
        } catch (IOException e) {
            throw new RuntimeException(e);
        } finally {
            try {
                xmlDocument.close();
            } catch (IOException e) {
                throw new RuntimeException(e);
            }

        }

    }
}