/*
 * Created on Feb 23, 2004
 */

package net.sourceforge.orbroker;

import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Member;
import java.lang.reflect.Method;
import java.net.URL;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.WeakHashMap;
import java.util.logging.ConsoleHandler;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.LogRecord;
import java.util.logging.Logger;

import javax.sql.DataSource;

import net.kildenpedersen.reflect.Reflector;
import net.kildenpedersen.xml.DOMBuilder;
import net.sourceforge.orbroker.exception.DefaultExceptionEvaluator;
import net.sourceforge.orbroker.exception.ExceptionEvaluator;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * The Broker between the JDBC data source and an application. The Broker holds
 * the configuration and initiates Queries and Transactions.
 * 
 * @author Nils Kilden-Pedersen
 * @version 2.0
 * @see net.sourceforge.orbroker.Query
 * @see net.sourceforge.orbroker.Transaction
 */
public final class Broker {
	
    private final class ConfigHandler {
    	private final static String ATTR_CATALOG = "catalog";
        private final static String ATTR_CLASS = "class";
        private final static String ATTR_DFT_LAZY = "lazy-load";
        private final static String ATTR_EXTENDS = "extends";
        private final static String ATTR_EXTERNAL = "external-source";
        private final static String ATTR_FROM_COLUMN = "from-column";
        private final static String ATTR_ID = "id";
        private final static String ATTR_ID_SUFFIX = "id-suffix";
        private final static String ATTR_KEY_COLUMNS = "key-columns";
        private final static String ATTR_LAZY = "lazy-load";
        private final static String ATTR_MAPPED_COLUMN = "mapped-column";
        private final static String ATTR_NAME = "name";
        private final static String ATTR_ON_VALUE = "on-value";
        private final static String ATTR_QUERY_COLUMN = "query-column";
        private final static String ATTR_RESULT_OBJECT = "result-object";
        private final static String ATTR_STATEMENT = "sql-statement";
        private final static String ATTR_TRANSACTION_ISOLATION = "transaction-isolation";
        private final static String ATTR_USE_COLUMN = "use-column";
        private final static String ELEM_ARGUMENT = "argument";
        private final static String ELEM_COLUMN = "column";
        private final static String ELEM_COLUMN_ALIAS = "column-alias";
        private final static String ELEM_CONSTRUCTOR = "constructor";
        private final static String ELEM_DELEGATE_MAPPING = "delegate-mapping";
        private final static String ELEM_DELEGATE_TO = "delegate-to";
        private final static String ELEM_FACTORY_METHOD = "factory-method";
        private final static String ELEM_FIELD = "field";
        private final static String ELEM_MAP_WITH = "map-with";
        private final static String ELEM_PARAMETER = "set-parameter";
        private final static String ELEM_PROPERTY = "property";
        private final static String ELEM_RESULT_OBJECT = "result-object";
        private final static String ELEM_SET_METHOD = "method";
        private final static String ELEM_STATEMENT = "sql-statement";
        private final static String ELEM_SUB_QUERY = "sub-query";

        private final DOMBuilder builder;
        private boolean lazyDefault;
        private final String schemaFile = "orbroker-2.0.xsd";
        private ConfigHandler() throws ConfigurationException {
            URL schemaURL = getClass().getResource("/" + this.schemaFile);
            if (schemaURL == null) {
                throw new ConfigurationException(this.schemaFile
                        + " not found.");
            }
            this.builder = new DOMBuilder(schemaURL);
        }

        private void addInstantiator(Element instantiatorElem,
                ResultObjectDefinition resultObjectDef) throws BrokerException {

        	String tagName = instantiatorElem.getTagName();
        	Class type = resultObjectDef.getType();
        	String methodName = instantiatorElem.getAttribute(ATTR_NAME);
        	
            NodeList parameters = instantiatorElem.getElementsByTagName(ELEM_ARGUMENT);
            Set candidates = 
            	getMethodOrConstructorCandidates(
            			tagName, type, parameters.getLength(), methodName);
            ArgumentList parmList = buildArgumentList(parameters, candidates);
            ResultObjectInstantiator instantiator = null;
            if (ELEM_CONSTRUCTOR.equals(tagName)) {
                instantiator = new ResultObjectConstructor(type, parmList);
            } else if (ELEM_FACTORY_METHOD.equals(tagName)) {
                instantiator = 
                	new ResultObjectFactoryMethod(type, methodName, parmList);
            } else {
                assert false : "Unknown element passed: "
                        + instantiatorElem.getTagName();
            }

            resultObjectDef.setInstantiator(instantiator);
        }

        private void addMappingDelegates(Element delegateMappingElem,
                ResultObjectDefinition resultObjectDef) throws BrokerException {

            String columnName = delegateMappingElem.getAttribute(ATTR_USE_COLUMN);
            Column discriminatorColumn = new Column(columnName, String.class);
            resultObjectDef.setDiscriminatorColumn(discriminatorColumn);

            NodeList delegateList = 
            	delegateMappingElem.getElementsByTagName(ELEM_DELEGATE_TO);
            for (int i = 0; i < delegateList.getLength(); i++) {
                Element delegateTo = (Element) delegateList.item(i);
                String resultObjectId = delegateTo
                        .getAttribute(ATTR_RESULT_OBJECT);
                if (resultObjectId.equals(resultObjectDef.getId())) {
                    String msg = "A " + delegateMappingElem.getTagName()
                            + " cannot reference itself, in this case: "
                            + resultObjectId;
                    throw new ConfigurationException(msg);
                }
                ResultObjectDefinition dynamicDef = 
                	Broker.this.findResultObjectDefinition(resultObjectId);
                String dynamicValue = delegateTo.getAttribute(ATTR_ON_VALUE);
                resultObjectDef.addDynamicResultObject(dynamicDef, dynamicValue);
            }
        }

        private void addSetField(Element fieldElem,
                ResultObjectDefinition resultObjectDef) throws BrokerException {

            String fieldName = fieldElem.getAttribute(ATTR_NAME);
            SetField setField = new SetField(resultObjectDef.getType(), fieldName);
            Element valueTypeElem = DOMBuilder.getFirstChildElement(fieldElem);
            setField.setValueType(buildValueType(valueTypeElem, setField.getFieldType()));
            resultObjectDef.addSetter(setField);
        }

        private void addSetMethod(Element setMethodElem,
                ResultObjectDefinition resultObjectDef) throws BrokerException {

            String methodName = setMethodElem.getAttribute(ATTR_NAME);
            Class type = resultObjectDef.getType();
            NodeList parameters = setMethodElem.getElementsByTagName(ELEM_ARGUMENT);
            Set candidates = 
            	getMethodOrConstructorCandidates(
            		setMethodElem.getTagName(), type, parameters.getLength(), methodName);
            ArgumentList parmList = buildArgumentList(parameters, candidates);
            SetMethod setMethod = new SetMethod(type, methodName, parmList);
            resultObjectDef.addSetter(setMethod);
        }

        private void addSetProperty(Element propertyElem,
                ResultObjectDefinition resultObjectDef) throws BrokerException {

            String propertyName = propertyElem.getAttribute(ATTR_NAME);
            SetProperty setProperty = 
            	new SetProperty(resultObjectDef.getType(), propertyName);
            Element valueTypeElem = DOMBuilder.getFirstChildElement(propertyElem);
            setProperty.setParameter(
            		buildValueType(valueTypeElem, setProperty.getPropertyType()));
            resultObjectDef.addSetter(setProperty);
        }
        
        /**
         * 
         * @param arguments
         * @param candidateSet Set of either methods or constructors 
         * which have the same number of arguments as the argument list.
         * @return argument list
         * @throws BrokerException
         */
        private ArgumentList buildArgumentList(NodeList arguments, Set candidateSet) 
        		throws BrokerException {

        	Class[] parmTypes = findParameterTypes(candidateSet, arguments);
            ArgumentList argumentList = new ArgumentList(arguments.getLength());
            for (int argumentPos = 0; argumentPos < arguments.getLength(); argumentPos++) {
                Element argument = (Element) arguments.item(argumentPos);
                Element valueTypeElem = DOMBuilder.getFirstChildElement(argument);
                argumentList.add(buildValueType(valueTypeElem, parmTypes[argumentPos]));
            }
            argumentList.close();
            return argumentList;
        }
        
        private Column buildColumn(Element columnElem, Class columnToClass,
                String idAttributeName) {
            String columnName = columnElem.getAttribute(idAttributeName);
            return new Column(columnName, columnToClass);
        }

        private MapWith buildMapWith(Element mapWithElem,
                Class requiredType) throws BrokerException {

            String resultObjID = mapWithElem.getAttribute(ATTR_RESULT_OBJECT);
            ResultObjectDefinition resultObjectDef = findResultObjectDefinition(resultObjID);
            if (!requiredType.isAssignableFrom(resultObjectDef.getType())) {
                String msg = "Use of result object '" + resultObjID
                        + "' is not of required type " + Reflector.getClassName(requiredType);
                throw new ConfigurationException(msg);
            }
            MapWith mapWith = new MapWith(resultObjectDef);
            NodeList aliases = mapWithElem.getElementsByTagName(ELEM_COLUMN_ALIAS);
            for (int i = 0; i < aliases.getLength(); i++) {
            	Element alias = (Element) aliases.item(i);
            	String queryColumn = alias.getAttribute(ATTR_QUERY_COLUMN);
            	String mappedColumn = alias.getAttribute(ATTR_MAPPED_COLUMN);
            	mapWith.addAlias(mappedColumn, queryColumn);
			}
            return mapWith;
        }

        /**
         * @param stm
         * @param id
         * @param extensionElem
         * @param extensionsMap
         * @return Extended id
         */
        private String buildStatementExtension(final String stm, final String id,
                final Element extensionElem, final Map extensionsMap) {

            String extensionId = id
                    + extensionElem.getAttribute(ATTR_ID_SUFFIX);
            StringBuffer extendedStm = (StringBuffer) extensionsMap
                    .get(extensionId);
            if (extendedStm == null) {
                extendedStm = new StringBuffer(stm);
                extensionsMap.put(extensionId, extendedStm);
            }
            NodeList textNodes = extensionElem.getChildNodes();
            for (int i = 0; i < textNodes.getLength(); i++) {
                Node node = textNodes.item(i);
                if (node.getNodeType() != Node.TEXT_NODE
                        && node.getNodeType() != Node.CDATA_SECTION_NODE) {
                    continue;
                }
                String value = node.getNodeValue();
                extendedStm.append(value);
            }
            return extensionId;
        }

        private void buildStatements(Element statementElem)
                throws BrokerException {
        	HashMap extensionStatements = new HashMap();
            HashMap extendedResultObjectDefs = new HashMap();
            StringBuffer stm = new StringBuffer();
            String id = statementElem.getAttribute(ATTR_ID);

            ResultObjectDefinition resultObjectDef = null;
            String resultObjId = statementElem.getAttribute(ATTR_RESULT_OBJECT);
            if (resultObjId != null && resultObjId.length() > 0) {
                resultObjectDef = findResultObjectDefinition(resultObjId);
            }

            final String externalSource = statementElem.getAttribute(ATTR_EXTERNAL);
            if (externalSource != null && externalSource.trim().length() > 0) {
                Statement externalStatement = null;
                externalStatement = 
                	Statement.newInstance(id, resultObjectDef, externalSource);
                registerStatement(externalStatement);
                return;
            }

            NodeList nodes = statementElem.getChildNodes();
            for (int i = 0; i < nodes.getLength(); i++) {
                Node node = nodes.item(i);
                if (node.getNodeType() == Node.TEXT_NODE
                        || node.getNodeType() == Node.CDATA_SECTION_NODE) {
                    String value = node.getNodeValue();
                    stm.append(value);
                    Iterator iter = extensionStatements.values().iterator();
                    while (iter.hasNext()) {
                        Object obj = iter.next();
                        StringBuffer extendedStm = (StringBuffer) obj;
                        extendedStm.append(value);
                    }
                } else if (node.getNodeType() == Node.ELEMENT_NODE) {
                	Element appendStatement = (Element) node;
                	String extendedId = buildStatementExtension(
                			stm.toString(), id, appendStatement, extensionStatements);
                	String extResultObjectId = appendStatement.getAttribute(ATTR_RESULT_OBJECT);
                	if (extResultObjectId != null && extResultObjectId.trim().length() > 0) {
                		extendedResultObjectDefs.put(extendedId, 
                				Broker.this.findResultObjectDefinition(extResultObjectId));
                	}

                }
            }

            Statement statement = 
                Statement.newInstance(id, stm.toString(), resultObjectDef);
            registerStatement(statement);
            Iterator extensions = extensionStatements.entrySet().iterator();
            while (extensions.hasNext()) {
                Map.Entry entry = (Map.Entry) extensions.next();
                String extendedId = (String) entry.getKey();
                StringBuffer extendedStm = (StringBuffer) entry.getValue();
                ResultObjectDefinition extendedDef = 
                	(ResultObjectDefinition) extendedResultObjectDefs.get(extendedId);
                if (extendedDef == null) {
                	extendedDef = resultObjectDef;
                }
                statement = Statement.newInstance(
                		extendedId, extendedStm .toString(), extendedDef);
                registerStatement(statement);
            }
        }

        private ValueType buildSubQuery(Element subQueryElem, Class requiredType)
                throws BrokerException {

            String statementID = subQueryElem.getAttribute(ATTR_STATEMENT);
            Statement statement = Broker.this.findStatement(statementID);
            if (statement == null) {
                String msg = ATTR_STATEMENT + " '" + statementID
                        + "' does not exist.";
                throw new ConfigurationException(msg);
            }

            boolean lazy;
            String lazyAttr = subQueryElem.getAttribute(ATTR_LAZY);
            if (lazyAttr != null && lazyAttr.trim().length() > 0) {
                lazy = Boolean.valueOf(lazyAttr).booleanValue();
            } else if (Collection.class.isAssignableFrom(requiredType)
                    && requiredType.isInterface()) {
                lazy = this.lazyDefault;
            } else {
                lazy = false;
            }

            SubQuery subQuery = new SubQuery(requiredType, statementID,
                    Broker.this, lazy);
            NodeList parameterList = subQueryElem.getElementsByTagName(ELEM_PARAMETER);

            for (int i = 0; i < parameterList.getLength(); i++) {
                Element parameterElem = (Element) parameterList.item(i);
                String parameterName = parameterElem.getAttribute(ATTR_NAME);
                ValueType parameter = 
                	buildColumn(parameterElem, Object.class, ATTR_FROM_COLUMN);
                subQuery.addParameterValue(parameterName, parameter);
            }

            return subQuery;
        }

        private ValueType buildValueType(Element valueTypeElem, Class type)
                throws BrokerException {
            String tagName = valueTypeElem.getTagName();
            if (ELEM_COLUMN.equals(tagName)) {
                return buildColumn(valueTypeElem, type, ATTR_NAME);
            } else if (ELEM_MAP_WITH.equals(tagName)) {
                return buildMapWith(valueTypeElem, type);
            } else if (ELEM_SUB_QUERY.equals(tagName)) {
                return buildSubQuery(valueTypeElem, type);
            } else {
                throw newInvalidElement(tagName);
            }
        }

        private void configure(InputStream configStream) throws BrokerException {
            Document configDoc = this.builder.parse(configStream);
            Element root = configDoc.getDocumentElement();
            Broker.this.brokerName = root.getAttribute(ATTR_NAME);
            Broker.this.catalog = root.getAttribute(ATTR_CATALOG);
            if (Broker.this.catalog != null && Broker.this.catalog.trim().length() == 0) {
            	Broker.this.catalog = null;
            }
            this.lazyDefault = !"false".equals(root.getAttribute(ATTR_DFT_LAZY));

            Broker.this.transactionIsolation = 
            	getTransactionIsolation(root.getAttribute(ATTR_TRANSACTION_ISOLATION));

            /*
             * Find all result-objects first. They need to be put in the map
             * first, so the sql-statements can include them.
             */
            NodeList list = configDoc.getElementsByTagName(ELEM_RESULT_OBJECT);
            for (int i = 0; i < list.getLength(); i++) {

                Element roElem = (Element) list.item(i);
                String roID = roElem.getAttribute(ATTR_ID);
                String roClassName = roElem.getAttribute(ATTR_CLASS);

                String extendsId = roElem.getAttribute(ATTR_EXTENDS);
                if (extendsId != null && extendsId.trim().length() == 0) {
                    extendsId = null;
                }
                
                ResultKeyDef pkDef = null;
                String pkColumns = roElem.getAttribute(ATTR_KEY_COLUMNS);
                if (pkColumns != null && pkColumns.trim().length() > 0) {
                    pkDef = new ResultKeyDef(pkColumns, roID);
                }

                Class roType = getClassByName(roClassName);
                ResultObjectDefinition resultObjectDef = new ResultObjectDefinition(
                        roID, roType, extendsId, pkDef, Broker.this);
                registerResultObject(resultObjectDef);
            }

            // Find all statements
            list = configDoc.getElementsByTagName(ELEM_STATEMENT);
            for (int i = 0; i < list.getLength(); i++) {
                Element statement = (Element) list.item(i);
                buildStatements(statement);
            }

            // Finish with the result objects
            LinkedList unfinished = new LinkedList();
            list = configDoc.getElementsByTagName(ELEM_RESULT_OBJECT);
            for (int i = 0; i < list.getLength(); i++) {
            	unfinished.add(list.item(i));
            }
            while (!unfinished.isEmpty()) {
                Element resultObjElem = (Element) unfinished.removeFirst();
                try {
					finishResultObject(resultObjElem);
				} catch (ParentNotFinishedException e) {
					unfinished.addLast(resultObjElem);
				}
            }
        }

        /**
         * 
         * @param candidateSet
         * @param arguments
         * @return The only candidate member
         */
        private Class[] findParameterTypes(Set candidateSet, NodeList arguments) {

            if (candidateSet.size() == 1) {
            	Member member = (Member) candidateSet.toArray()[0];
            	return Reflector.getParameterTypes(member);
            }
            
        	assert candidateSet.size() > 0 : "List of candidate members is 0.";
        	
        	String name; {
	        	Member example = (Member) candidateSet.toArray()[0];
	        	if (example.getClass() == Method.class) {
	        		name = "method " + example.getName(); 
	        	} else {
	        		name = "constructor " + example.getDeclaringClass().getName();
	        	}
        	}
        	
        	Class[] parmTypes = new Class[arguments.getLength()];
        	
            for (int i = 0; i < arguments.getLength(); i++) {
                Element argument = (Element) arguments.item(i);
                Element childElem = DOMBuilder.getFirstChildElement(argument);
                String argClassName = argument.getAttribute(ATTR_CLASS);
                
                if (argClassName != null && argClassName.trim().length() > 0) {
                	parmTypes[i] = getClassByName(argClassName); 
                } else if (ELEM_MAP_WITH.equals(childElem.getTagName())) {
                	String resultObjectId = childElem.getAttribute(ATTR_RESULT_OBJECT);
                	if (resultObjectId != null && resultObjectId.trim().length() > 0) {
                		parmTypes[i] = findResultObjectDefinition(resultObjectId).getType();
                	}
                } else { 
                	parmTypes[i] = getArgumentTypeIfAllEqual(candidateSet, i);
                }

                if (parmTypes[i] != null) {
                	removeNonCompliant(candidateSet, i, parmTypes[i]);
                }

            }
            
            if (candidateSet.size() == 1) {
            	Member member = (Member) candidateSet.toArray()[0];
            	return Reflector.getParameterTypes(member);
            }

            StringBuffer msg = new StringBuffer();
            if (candidateSet.isEmpty()) {
            	msg.append("Cannot find ");
            } else {
            	msg.append("More than one ");
            }
            msg.append(name).append("(");
        	for (int i = 0; i < parmTypes.length; i++) {
				if (i > 0) {
					msg.append(", ");
				}
				msg.append(parmTypes[i] != null ? parmTypes[i].getName() : "?");
			}
        	msg.append(").");
        	if (!candidateSet.isEmpty()) {
        		msg.append(" Please use attribute '").append(ATTR_CLASS).append("' to narrow.");
        	}
           	throw new ConfigurationException(msg.toString());
        }

        /**
         * Finish building the {@link ResultObjectDefinition}. All result objects
         * have already been constructed to avoid unresolved dependencies. So
         * they must be finished by retrieving them from the Map.
         * @param resultObjElem
         * @throws BrokerException
         * @throws ParentNotFinishedException
         */
        private void finishResultObject(Element resultObjElem)
                throws BrokerException, ParentNotFinishedException {
            String roID = resultObjElem.getAttribute(ATTR_ID);
            ResultObjectDefinition resultObjectDef = findResultObjectDefinition(roID);

            // Initialize with parent, if exist
            String extendsId = resultObjectDef.getExtendsId();
            if (extendsId != null) {
            	if (!Broker.this.resultObjectFactories.containsKey(extendsId)) {
                    String msg = ELEM_RESULT_OBJECT + " '" + roID
                            + "' extends '" + extendsId + "', but '"
                            + extendsId + "' is not defined.";
                    throw new ConfigurationException(msg);
                }
            	ResultObjectDefinition parentDef = findResultObjectDefinition(extendsId);
            	if (!parentDef.isClosed()) {
            		throw new ParentNotFinishedException();
            	}
                resultObjectDef.initializeWith(parentDef);
            }

            NodeList roNodes = resultObjElem.getChildNodes();
            for (int i = 0; i < roNodes.getLength(); i++) {
                Node node = roNodes.item(i);
                if (node.getNodeType() != Node.ELEMENT_NODE) {
                    continue;
                }
                Element genericElem = (Element) node;
                String tagName = genericElem.getTagName();
                if (ELEM_DELEGATE_MAPPING.equals(tagName)) {
                    addMappingDelegates(genericElem, resultObjectDef);
                } else if (ELEM_CONSTRUCTOR.equals(tagName)) {
                    addInstantiator(genericElem, resultObjectDef);
                } else if (ELEM_FACTORY_METHOD.equals(tagName)) {
                    addInstantiator(genericElem, resultObjectDef);
                } else if (ELEM_PROPERTY.equals(tagName)) {
                    addSetProperty(genericElem, resultObjectDef);
                } else if (ELEM_SET_METHOD.equals(tagName)) {
                    addSetMethod(genericElem, resultObjectDef);
                } else if (ELEM_FIELD.equals(tagName)) {
                    addSetField(genericElem, resultObjectDef);
                } else {
                    throw new ConfigurationException("Invalid element '"
                            + tagName + "'.");
                }
            }

            // Set no-arg constructor, if none exist and not abstract/interface
            if (!resultObjectDef.hasInstantiator()
                    && resultObjectDef.isInstantiable()) {
                ResultObjectInstantiator instantiator = new ResultObjectConstructor(
                        resultObjectDef.getType(), ArgumentList.NO_ARG);
                resultObjectDef.setInstantiator(instantiator);
            }
            resultObjectDef.close();
        }

        /**
         * Return the argument type at the given position if it's the same
         * for all members, or <code>null</code> if it's different.
         * @param members
         * @param argumentPos
         * @return type, or null if disagreement
         */
        private Class getArgumentTypeIfAllEqual(Set members, int argumentPos) {
        	Class returnType = null;
        	Iterator iter = members.iterator();
        	while (iter.hasNext()) {
        		Class type = Reflector.getParameterTypes((Member) iter.next())[argumentPos];
        		if (returnType == null) {
        			returnType = type;
        		} else if (returnType != type) {
        			return null;
        		}
			}
        	return returnType;
        }
        
        private Class getClassByName(final String className)
                throws ConfigurationException {
            Class cls = (Class) Broker.PRIMITIVE_CLASSES.get(className);
            if (cls != null) {
                return cls;
            }
            try {
                return Class.forName(className);
            } catch (ClassNotFoundException e) {
                throw new ConfigurationException(e);
            }
        }
        private Set getMethodOrConstructorCandidates(final String tagName, final Class type, final int parmCount, String methodName) {
            Set members = null;
            
            if (ELEM_CONSTRUCTOR.equals(tagName)) {
                members = Reflector.getAccessibleConstructors(type, parmCount);
                methodName = null;
            } else if (ELEM_SET_METHOD.equals(tagName)) {
                members = Reflector.getAccessibleMemberMethods(type, methodName, parmCount);
            } else if (ELEM_FACTORY_METHOD.equals(tagName)) {
                members = Reflector.getAccessibleStaticMethods(type, methodName, parmCount);
            } else {
                assert false : "Unknown element passed: " + tagName;
            }
            
            if (members == null || members.size() == 0) {
                String msg = 
                    "No " + tagName + 
                    (methodName != null ? " named '" + methodName + "' " : "") +
                    " with " + parmCount + " argument(s) exist for class '" + Reflector.getClassName(type) + "'"; 
                throw new ConfigurationException(msg);
            }
            
            return members;
        }
        
        /**
         * 
         * @param isolation
         * @return transaction isolation level as Integer, or <code>null</code>
         *              if it doesn't exist
         */
        private Integer getTransactionIsolation(String isolation) {
            if (isolation == null) {
                return null;
            }
            if (isolation.trim().length() == 0) {
                return null;
            }
            Integer isolationLevel = 
            	(Integer) Broker.TRANSACTION_ISOLATIONS.get(isolation);
            if (isolationLevel == null) {
                String msg = "Unknown transaction isolation level: "
                        + isolation;
                throw new ConfigurationException(msg);
            }
            return isolationLevel;
        }

        private ConfigurationException newInvalidElement(String tagName) {
            String msg = "Invalid element <" + tagName + "> encountered.";
            return new ConfigurationException(msg);
        }
        
        private void registerResultObject(ResultObjectDefinition resultObjectDef) 
        		throws ConfigurationException {
		    String id = resultObjectDef.getId();
		    if (Broker.this.resultObjectFactories.containsKey(id)) {
		        String msg = ELEM_RESULT_OBJECT + " '" + id
		                + "' defined multiple times in Broker '"
		                + Broker.this.getName() + "'";
		        throw new ConfigurationException(msg);
		    }
		    Broker.this.resultObjectFactories.put(id, resultObjectDef);
		    Broker.log(Level.CONFIG, "Result-object registrered: " + id);
		}

        private void registerStatement(Statement statement)
                throws ConfigurationException {
            String id = statement.getId();
            if (Broker.this.sqlStatements.containsKey(id)) {
                String msg = ELEM_STATEMENT + " '" + id
                        + "' defined multiple times in Broker '"
                        + Broker.this.getName() + "'";

                throw new ConfigurationException(msg);
            }
            Broker.this.sqlStatements.put(id, statement);
            Broker.log(Level.CONFIG, "Statement registered: " + id);
        }
        
        private void removeNonCompliant(
        		Set memberCandidates, 
				int argumentPos, 
				Class mustBeType) {
        	
        	Iterator iter = memberCandidates.iterator();
        	while (iter.hasNext()) {
        		Member member = (Member) iter.next();
        		if (Reflector.getParameterTypes(member)[argumentPos] != mustBeType) {
        			iter.remove();
        		}
        	}
        }
    }

	private final class ParentNotFinishedException extends Exception {
		// Nothing to see here. Move along.
	}
	
    private static final WeakHashMap cachedObjects = new WeakHashMap();
    private final static Handler DEFAULT_LOG_HANDLER = new ConsoleHandler();
    private final static Logger LOGGER = Logger.getLogger(Broker.class.getName());
    private final static HashMap PRIMITIVE_CLASSES = new HashMap();
    private static final HashMap TRANSACTION_ISOLATIONS = new HashMap();
    static {
        TRANSACTION_ISOLATIONS.put("TRANSACTION_NONE", new Integer(
                Connection.TRANSACTION_NONE));
        TRANSACTION_ISOLATIONS.put("TRANSACTION_READ_UNCOMMITTED", new Integer(
                Connection.TRANSACTION_READ_UNCOMMITTED));
        TRANSACTION_ISOLATIONS.put("TRANSACTION_READ_COMMITTED", new Integer(
                Connection.TRANSACTION_READ_COMMITTED));
        TRANSACTION_ISOLATIONS.put("TRANSACTION_REPEATABLE_READ", new Integer(
                Connection.TRANSACTION_REPEATABLE_READ));
        TRANSACTION_ISOLATIONS.put("TRANSACTION_SERIALIZABLE", new Integer(
                Connection.TRANSACTION_SERIALIZABLE));

        Map switched = new HashMap(TRANSACTION_ISOLATIONS.size());
        Iterator entries = TRANSACTION_ISOLATIONS.entrySet().iterator();
        while (entries.hasNext()) {
        	Map.Entry entry = (Map.Entry) entries.next();
        	switched.put(entry.getValue(), entry.getKey());
        }
        TRANSACTION_ISOLATIONS.putAll(switched);
    }
    
    static {
        LOGGER.setUseParentHandlers(false);
        DEFAULT_LOG_HANDLER.setFormatter(new BrokerFormatter());
        DEFAULT_LOG_HANDLER.setLevel(Level.ALL);
        LOGGER.addHandler(DEFAULT_LOG_HANDLER);
    }
    static {
    	PRIMITIVE_CLASSES.put("object", Object.class);
        PRIMITIVE_CLASSES.put("int", int.class);
        PRIMITIVE_CLASSES.put("string", String.class);
        PRIMITIVE_CLASSES.put("boolean", boolean.class);
        PRIMITIVE_CLASSES.put("long", long.class);
        PRIMITIVE_CLASSES.put("short", short.class);
        PRIMITIVE_CLASSES.put("byte", byte.class);
        PRIMITIVE_CLASSES.put("char", char.class);
        PRIMITIVE_CLASSES.put("double", double.class);
        PRIMITIVE_CLASSES.put("float", float.class);
        PRIMITIVE_CLASSES.put("byte[]", byte[].class);
        PRIMITIVE_CLASSES.put("char[]", char[].class);
    }

    static Map getCachedObjects(Connection con) {
    	synchronized (con) {
    		Map map = (Map) cachedObjects.get(con);
    		if (map == null) {
    			map = new HashMap();
    			cachedObjects.put(con, map);
    		}
    		return map;
    	}
    }

    static String getTransactionIsolationLevel(int isolationLevel) {
    	return (String) TRANSACTION_ISOLATIONS.get(new Integer(isolationLevel));
    }

	/**
	 * @param scrollableSupport
	 * @return result set type
	 */
	static int getResultSetType(boolean scrollableSupport) {
		if (scrollableSupport) {
			return ResultSet.TYPE_SCROLL_SENSITIVE;
		}
		return ResultSet.TYPE_FORWARD_ONLY;
	}

    static boolean isLoggable(Level level) {
        return LOGGER.isLoggable(level);
    }

    /**
     * Add logging handler. When adding a logging handler, the default console
     * log handler is automatically removed. Previously added handlers remain.
     * 
     * @param handler Logging handler.
     * @see Handler
     * @see #addLoggingHandler(Handler, boolean)
     */
    public static void addLoggingHandler(Handler handler) {
    	addLoggingHandler(handler, false);
    }

    /**
     * Add logging handler. Previously added handlers remain.
     * @param handler Logging handler.
     * @param keepDefaultHandler true, if default console log handler should remain
     * @see Handler
     * @see #addLoggingHandler(Handler)
     */
    public static void addLoggingHandler(Handler handler, boolean keepDefaultHandler) {
        LOGGER.addHandler(handler);
        if (!keepDefaultHandler) {
        	LOGGER.removeHandler(DEFAULT_LOG_HANDLER);
        }
    }
    
    /**
     * Log a message.
     * 
     * @param level Log level
     * @param message Log message
     */
    public static void log(Level level, String message) {
        if (LOGGER.isLoggable(level)) {
            LOGGER.log(new LogRecord(level, message));
        }
    }

    /**
     * Set the logging level. Log level definitions:
     * <ul>
     *		<li>WARNING: Used for internally caught exceptions not considered severe.</li>
     *		<li>INFO: Shows transaction boundaries and SQL statements executed.</li>
     *		<li>CONFIG: Information about driver support and configuration setup</li>
     *		<li>FINE: Shows dynamic SQL statement after parsing.</li>
     * </ul>
     * 
     * @param level Logging level
     */
    public static void setLoggingLevel(Level level) {
        LOGGER.setLevel(level);
    }

    /**
     * Set the logging level. See log level definitions on {@link #setLoggingLevel(Level)}.
     * @param level Logging level
     * @see Level
     */
    public static void setLoggingLevel(String level) {
        setLoggingLevel(Level.parse(level));
    }
    private String brokerName;
    private String catalog;
    private DataSource dataSource = null;
    private ExceptionEvaluator evaluator = new DefaultExceptionEvaluator();
    private final WeakHashMap executableConnections = new WeakHashMap();
    private char[] password = null;
    private final HashMap resultObjectFactories = new HashMap();
    private final HashMap sqlStatements = new HashMap();
    private boolean supportsBatch = false;
    private boolean supportsScrollable = false;
    private TextReplacements textReplacements = new TextReplacements();
    private Integer transactionIsolation;
    private String username = null;

    /**
     * Create Broker without a configuration.
     * 
     * @param dataSource A data source
     * @see #addStatement(String, String)
     */
    public Broker(DataSource dataSource) {
        this(dataSource, null, null);
    }

    /**
     * Create Broker without a configuration.
     * 
     * @param dataSource A data source
     * @param username Username to use that is different than on data source
     * @param password Password for username
     * @see #addStatement(String, String)
     */
    public Broker(DataSource dataSource, String username, String password) {
        setDataSource(dataSource, username, password);
    }

    /**
     * The Broker is configured by passing an {@link InputStream} containing an
     * XML file and a data source.
     * 
     * @param configuration The XML configuration
     * @param dataSource A data source
     * @throws BrokerException
     */
    public Broker(InputStream configuration, DataSource dataSource)
            throws BrokerException {
        this(configuration, dataSource, null, null);
    }

    /**
     * The Broker is configured by passing an {@link InputStream} containing an
     * XML file and a data source with username and password.
     * 
     * @param configuration The XML configuration
     * @param dataSource A data source
     * @param username Username to use that is different than on data source
     * @param password Password for username
     * @throws BrokerException
     */
    public Broker(InputStream configuration, DataSource dataSource,
            String username, String password) throws BrokerException {
        ConfigHandler handler = new ConfigHandler();
        handler.configure(configuration);
        try {
            configuration.close();
        } catch (IOException e) {
            Broker.log(Level.WARNING, e.toString());
        }
        setDataSource(dataSource, username, password);
    }
    
    private void checkMetaData() {
        Connection con = null;
        try {
            con = getFreshConnection(null);
            DatabaseMetaData metaData = con.getMetaData();
            this.supportsScrollable = metaData.supportsResultSetConcurrency(
            		getResultSetType(true),
                    ResultSet.CONCUR_READ_ONLY);
            this.supportsBatch = metaData.supportsBatchUpdates();
        } catch (SQLException e) {
            log(Level.WARNING, e.toString());
        } finally {
            if (con != null) {
                try {
                    con.close();
                } catch (SQLException e) {
                    log(Level.WARNING, e.toString());
                }
            }
        }

        if (!this.supportsScrollable) {
            log(Level.CONFIG, "Scrollable result sets not supported.");
        }
        if (!this.supportsBatch) {
            log(Level.CONFIG, "Batch updates not supported.");
        }

    }

    ResultObjectDefinition findResultObjectDefinition(String resultObjectID)
            throws BrokerException {
        ResultObjectDefinition resultObjectDef = (ResultObjectDefinition) this.resultObjectFactories
                .get(resultObjectID);
        if (resultObjectDef == null) {
            throw new BrokerException("Unknown result-object id: "
                    + resultObjectID);
        }
        return resultObjectDef;
    }

    Statement findStatement(String statementID) throws BrokerException {
        Statement stm = (Statement) this.sqlStatements.get(statementID);
        if (stm == null) {
            String msg = "Statement '" + statementID + "' unknown to Broker '"
                    + this.brokerName + "'";
            throw new BrokerException(msg);
        }
        return stm;
    }

    ExceptionEvaluator getExceptionEvaluator() {
        return this.evaluator;
    }

    /**
     * Get fresh connection from data source. 
     * 
     * @param isolationLevel
     * @return Fresh connection
     */
    Connection getFreshConnection(Integer isolationLevel) {
        Connection con;
        try {

        	if (this.username != null && this.password != null) {
                con = this.dataSource.getConnection(this.username, String.valueOf(this.password));
            } else {
                con = this.dataSource.getConnection();
            }

            if (con == null) {
                throw new BrokerException(
                        "Broker '" + this.brokerName + "' unable to obtain connection from data source.");
            }

            if (this.catalog != null) {
            	con.setCatalog(this.catalog);
            }

            if (isolationLevel != null
                    && con.getTransactionIsolation() != isolationLevel.intValue()) {
                con.setTransactionIsolation(isolationLevel.intValue());
            }

            boolean autoCommit = (con.getTransactionIsolation() == Connection.TRANSACTION_NONE);
            if (con.getAutoCommit() != autoCommit) {
            	con.setAutoCommit(autoCommit);
            }
            
        } catch (SQLException e) {
            throw new BrokerException(e.toString());
        }
        
        return con;
    }

    /**
     * @return Returns the text replacements.
     */
    TextReplacements getTextReplacements() {
        return this.textReplacements;
    }

	boolean supportsBatch() {
        return this.supportsBatch;
    }

    /**
     * @return Returns scrollable support.
     */
    boolean supportsScrollable() {
        return this.supportsScrollable;
    }

    /**
     * Add statement to Broker.
     * 
     * @param id The statement id.
     * @param sql The SQL statement
     * @throws BrokerException
     * @throws ConfigurationException
     */
    public void addStatement(final String id, final String sql)
            throws BrokerException, ConfigurationException {
        addStatement(id, sql, null);
    }

    /**
     * Add statement to Broker.
     * 
     * @param id The statement id.
     * @param sql The SQL statement
     * @param resultObjectId The <code>result-object</code> id as defined in the XML
     *                   configuration file.
     * @throws BrokerException
     * @throws ConfigurationException
     */
    public void addStatement(final String id, final String sql,
            String resultObjectId) throws BrokerException,
            ConfigurationException {

        ResultObjectDefinition resultObjectDef = null;
        if (resultObjectId != null) {
            resultObjectDef = findResultObjectDefinition(resultObjectId);
        }
        Statement stm = Statement.newInstance(id, sql, resultObjectDef);
        if (this.sqlStatements.containsKey(id)) {
            throw new ConfigurationException("Statement '" + id
                    + "' already exists.");
        }
        this.sqlStatements.put(id, stm);
    }

    /**
     * Get the data source used by Broker.
     * @return The datasource used.
     */
    public DataSource getDataSource() {
        return this.dataSource;
    }

    /**
     * Get brokerName of Broker.
     * @return Name of broker
     */
    public String getName() {
        return this.brokerName;
    }

    /**
     * Obtain an Executable. An Executable instance <i>must</i> be
     * released again by calling {@link #releaseExecutable(Executable)},
     * or unpredictable behavior can occur.
     * @param connection An externally managed connection.
     * @return Executable
     * @see #releaseExecutable(Executable)
     * @see Transaction#obtainExecutable()
     */
    public Executable obtainExecutable(Connection connection) {
    	Executable exe = new Executable(this, connection); 
    	this.executableConnections.put(exe, connection);
    	return exe;
    }

    /**
     * Release Executable. This will drop the internal Connection and prevent
     * other processes, which may hold a reference to the Executable, from
     * executing statements on a Connection that may have been returned to a
     * DataSource connection pool. 
     * <p><i>Releasing the Executable will not close the connection.</i></p>
     * @param executable The Executable to release
     * @throws BrokerException if the Executable was not obtained from this Broker
     * @see #obtainExecutable(Connection)
     */
    public void releaseExecutable(Executable executable) {
    	executable.release();
    	Connection con = (Connection) this.executableConnections.remove(executable);
    	if (con == null) {
    		String msg = 
    			"Executable was not obtained from Broker '" + this.brokerName + "'";
    		throw new BrokerException(msg);
    	}
    	getCachedObjects(con).clear();
    }
	
	/**
	 * Set the catalog.
	 * @param catalog The catalog name
	 */
	public void setCatalog(String catalog) {
		this.catalog = catalog;
	}

    /**
     * Set the data source. Uses default username and password.
     * @param dataSource The data source
     */
    public void setDataSource(DataSource dataSource) {
        setDataSource(dataSource, null, null);
    }
    
    /**
     * Set data source and username/password.
     * 
     * @param dataSource The data source
     * @param username Username to use that is different than on data source
     * @param password Password for username
     */
    public void setDataSource(DataSource dataSource, String username,
            String password) {

        if (dataSource == null) {
            throw new NullPointerException("Data source is null.");
        }
        this.dataSource = dataSource;
        this.username = username;
        if (password != null) {
            this.password = password.toCharArray();
        } else {
            this.password = null;
        }

        checkMetaData();
    }
    
    /**
     * Set a new exception evaluator.
     * @param evaluator The exception evaluator instance
     */
    public void setExceptionEvaluator(ExceptionEvaluator evaluator) {
        if (evaluator == null) {
            throw new NullPointerException("Exception evaluator is null.");
        }
        this.evaluator = evaluator;
    }

    /**
     * Set a text replacement value. This will replace a <code>{{key}}</code>
     * string with the specific parameter value.
     * @param key Text replacement key
     * @param value Text replacement value
     * @see #setTextReplacements(Properties)
     */
    public void setTextReplacement(String key, String value) {
        this.textReplacements.setProperty(key, value);
    }

    /**
     * Set text replacement values. This will replace all <code>{{key}}</code>
     * type properties in an sql-statement with the values.
     * 
     * @param textReplacements The replacement values to set.
     * @see #setTextReplacement(String, String)
     */
    public void setTextReplacements(Properties textReplacements) {
        this.textReplacements.setAll(textReplacements);
    }

    /**
     * Start a read-only query.
     * 
     * @return Query
     */
    public Query startQuery() {
        return new Query(this, this.transactionIsolation);
    }

    /**
     * Start a read-only query with a given transaction isolation level.
     * JDBC transaction isolation levels:
     * <ul>
     * 	<li>{@link Connection#TRANSACTION_NONE}</li>
     * 	<li>{@link Connection#TRANSACTION_READ_UNCOMMITTED}</li>
     * 	<li>{@link Connection#TRANSACTION_READ_COMMITTED}</li>
     * 	<li>{@link Connection#TRANSACTION_REPEATABLE_READ}</li>
     * 	<li>{@link Connection#TRANSACTION_SERIALIZABLE}</li> 
     * </ul>
     * @param isolationLevel The isolation level
     * @return Query
     */
    public Query startQuery(int isolationLevel) {
        return new Query(this, new Integer(isolationLevel));
    }

    /**
     * Start a new transaction.
     * 
     * @return Transaction
     * @throws BrokerException
     */
    public Transaction startTransaction() throws BrokerException {
        return new Transaction(this, this.transactionIsolation);
    }

    /**
     * Start a new transaction with a given transaction isolation level.
     * JDBC transaction isolation levels:
     * <ul>
     * 	<li>{@link Connection#TRANSACTION_NONE}</li>
     * 	<li>{@link Connection#TRANSACTION_READ_UNCOMMITTED}</li>
     * 	<li>{@link Connection#TRANSACTION_READ_COMMITTED}</li>
     * 	<li>{@link Connection#TRANSACTION_REPEATABLE_READ}</li>
     * 	<li>{@link Connection#TRANSACTION_SERIALIZABLE}</li> 
     * </ul>

     * @param isolationLevel The isolation level
     * @return Transaction
     * @throws BrokerException
     */
    public Transaction startTransaction(int isolationLevel)
            throws BrokerException {
        return new Transaction(this, new Integer(isolationLevel));
    }
    
    /**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "Broker: " + this.brokerName;
	}
	
}
