/*
 * Created on Jul 10, 2004
 */
package net.sourceforge.orbroker;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.logging.Level;

import net.sourceforge.orbroker.exception.ExceptionEvaluator;


/**
 * An abstract connection wrapper.
 * @author Nils Kilden-Pedersen
 */
abstract class BrokerConnection {
	private final boolean autoCommit;
	
	private final Broker broker;
	private Connection connection;
	private final ConnectionContext context = new ConnectionContext();
	private final int isolationLevel;
	private Map localStatements = null;
	private boolean retryStaleConnection;
	
	private BrokerConnection(
	        Broker broker, 
	        Connection connection, 
	        boolean retryStaleConnection) {
		this.context.setTextReplacements(broker.getTextReplacements());
		this.broker = broker;
	    this.connection = connection;
	    this.retryStaleConnection = retryStaleConnection; 
	    try {
            setConnectionProperties();
            this.autoCommit = this.connection.getAutoCommit();
    	    this.isolationLevel = this.connection.getTransactionIsolation();
        } catch (SQLException e) {
            throw new BrokerException(e);
        }
	}
	
	/**
	 * Create BrokerConnection with external JDBC connection. 
	 * Will not retry a stale connection.
	 * @param broker
	 * @param connection
	 */
	protected BrokerConnection(Broker broker, Connection connection) {
	    this(broker, connection, false);
	}
	
	/**
	 * Create BrokerConnection with a Broker connection. 
	 * Will retry a stale connection. 
	 * @param broker
	 * @param isolationLevel
	 */
	protected BrokerConnection(Broker broker, Integer isolationLevel) {
	    this(broker, broker.getFreshConnection(isolationLevel), true);
	}
	
	private final void addLocalStatement(Statement statement) {
	    if (this.localStatements == null) {
	        this.localStatements = new HashMap();
	    }
	    this.localStatements.put(statement.getId(), statement);
	}
	
	/**
	 * Release connection, and close if necessary.
	 */
	protected final void dropConnection() {
	    
		if (this.connection == null) {
			return;
		}
		
		try {
			if (shouldConnectionClose()) {
				Broker.getCachedObjects(this.connection).clear();
				if (!this.connection.isClosed()) {
					this.connection.close();
				}
			}
		} catch (SQLException e) {
		    Broker.log(Level.WARNING, e.toString());
		} finally {
			this.connection = null;
		}
	}
	
	/**
	 * Finalize.
	 * @see java.lang.Object#finalize()
	 */
	protected final void finalize() {
		if (this.connection != null) {
			dropConnection();
			Broker.log(Level.WARNING, getFinalizeWarning());
		}
		try {
			super.finalize();
		} catch (Throwable e) {
			Broker.log(Level.WARNING, e.toString());
		}
	}
	
	protected final Connection getActiveConnection() {
		if (this.connection == null) {
			throw new IllegalStateException(getClosedConnectionMessage());
		}
		return this.connection;
	}
	
	protected abstract String getClosedConnectionMessage();
	
	protected final ConnectionContext getConnectionContext() {
	    return this.context;
	}
	
	protected abstract String getFinalizeWarning();
	
	/**
	 * Return the statement. Give preference to any locally defined statements.
	 * @param statementID
	 * @return The statement with the given id
	 * @throws BrokerException
	 */
	protected Statement getStatement(String statementID) throws BrokerException {
		if (this.localStatements != null) {
			Statement localStatement = (Statement) this.localStatements.get(statementID);
			if (localStatement != null) {
				return localStatement;
			}
		}
		return this.broker.findStatement(statementID);
	}
	
	protected final String getTransactionIsolationText() {
		return Broker.getTransactionIsolationLevel(this.isolationLevel);
	}
	
	protected final boolean isAutoCommit() {
		return this.autoCommit;
	}

	protected final Executable newExecutable(Transaction transaction) {
	    return new Executable(this.broker, transaction);
	}

	/**
	 * Retry is guaranteed to return true 0 or 1 times only, 
	 * so it can be used recursively without looping infinitely.
	 * @param e exception
	 * @return true if statement should be retried
	 */
	protected boolean retryFailedStatement(SQLException e) {
		if (this.retryStaleConnection) {
		    ExceptionEvaluator evaluator = this.broker.getExceptionEvaluator(); 
		    if (evaluator.isStaleConnection(e)) {
		        this.connection = this.broker.getFreshConnection(new Integer(this.isolationLevel));
		        this.retryStaleConnection = false;
		        return true;
		    }
		}
	    return false;
	}

	protected abstract void setConnectionProperties() throws SQLException;
	
	protected abstract boolean shouldConnectionClose();

	protected final boolean supportsBatch() {
		return this.broker.supportsBatch();
	}
	
	protected final boolean supportsScrollable() {
		return this.broker.supportsScrollable();
	}
	
    /**
     * Determine exception type and throw exception.
     * @param source SQLException
     * @throws BrokerException
     * @throws ConstraintException
     * @throws DeadlockException
     */
    protected final void throwException(SQLException source) 
    		throws BrokerException, ConstraintException, DeadlockException {

        ExceptionEvaluator evaluator = this.broker.getExceptionEvaluator(); 
        if (evaluator.isConstraint(source)) {
            throw new ConstraintException(source.getMessage(), source);
        }

        if (evaluator.isDeadlock(source)) {
            throw new DeadlockException(source.getMessage(), source);
        }

    	throw new BrokerException(source);
    }

	/**
	 * Add context parameters.
	 * @param addContext The context to add.
	 */
	final void addContext(ConnectionContext addContext) {
	    this.context.addContext(addContext);
	}
	
	/**
	 * Add temporary statement.
	 * @param id The statement id.
	 * @param sql The SQL statement
	 * @throws BrokerException
	 * @throws ConfigurationException
	 */
	public final void addStatement(String id, String sql)  
			throws BrokerException, ConfigurationException {
		addStatement(id, sql, null);
	}

	/**
	 * Add temporary statement.
	 * @param id The statement id.
	 * @param sql The SQL statement
	 * @param resultObjectId The <code>result-object</code> id as 
	 * defined in the XML configuration file. 
	 * @throws BrokerException
	 * @throws ConfigurationException
	 */
	public final void addStatement(
			String id, 
			String sql, 
			String resultObjectId)
			throws BrokerException, ConfigurationException {
		ResultObjectDefinition resultObjectDef = null;
		if (resultObjectId != null) {
			resultObjectDef = this.broker.findResultObjectDefinition(resultObjectId);
		}
		
		Statement stm = 
		    Statement.newInstance(id, sql, resultObjectDef);
		addLocalStatement(stm);
	}
	
	/**
	 * Get parameter.
	 * Used to retrieve OUT/INOUT parameters from stored procedures.
	 * Both input and output parameters are stored in the same context,
	 * so any parameter set using {@link #setParameter(String, Object)}
	 * will be available with this method. Also if a stored procedure has
	 * an output parameter that needs to be used as input in another 
	 * SQL statements, it's immediately available, i.e. there's <i>no need</i>
	 * to code like this: 
	 * <code style="text-decoration: line-through;">
	 * txn.setParameter("parm", txn.getParameter("parm"));</code>   
	 * @param name parameter name
	 * @return parameter value
	 */
	public final Object getParameter(String name) {
	    return this.context.getParameterValue(name);
	}

	/**
	 * Set named parameter.
	 * @param name parameter name
	 * @param value parameter value
	 */
	public final void setParameter(String name, Object value) {
	    this.context.setParameter(name, value);
	}
	
	/**
	 * Set a text replacement value. This will replace a <code>{{key}}</code> 
	 * string with the value. 
	 * This will override any value set on Broker.
	 * @param key Text replacement key
	 * @param value Text replacement value
	 * @see #setTextReplacements(Properties)
	 */
	public final void setTextReplacement(String key, String value) {
	    this.context.setTextReplacement(key, value);
	}

	/**
	 * Set text replacement values. This will replace all <code>{{key}}</code> 
	 * type properties in an sql-statement with the values. 
	 * This will override any value set on Broker.
	 * @param textReplacements The replacement values to set.
	 * @see #setTextReplacement(String, String)
	 */
	public final void setTextReplacements(Properties textReplacements) {
	    this.context.setTextReplacements(textReplacements);
	}

}

