/*
 * Created on Feb 22, 2004
 */
package net.sourceforge.orbroker;

import java.sql.Connection;
import java.sql.SQLException;

/**
 * <p>An Executable allows queries and execution of statements
 * on an existing connection. It is up to the Connection supplier to
 * manage all transactional properties, if any. This should generally 
 * only be used for container managed transactions and 
 * similar type external transaction management.
 * </p> 
 * <p>An Executable can also be passed to other methods that 
 * needs to take part in a transaction, but without exposing the 
 * commit and rollback capabilities of the 
 * {@link net.sourceforge.orbroker.Transaction} object.
 * </p>  
 * 
 * @author Nils Kilden-Pedersen
 * @see Broker#obtainExecutable(Connection)
 * @see Transaction#obtainExecutable()
 */
public final class Executable extends ExecutableConnection {
	
	private final Transaction transaction;
	
	Executable(Broker owner, Connection connection) throws BrokerException {
		super(owner, connection);
		this.transaction = null;
	}

	Executable(Broker owner, Transaction transaction) throws BrokerException {
		super(owner, transaction.getActiveConnection());
		this.transaction = transaction;
	}

	/**
	 * @see net.sourceforge.orbroker.QueryableConnection#getClosedConnectionMessage()
	 */
	protected String getClosedConnectionMessage() {
		return "Executable has been released and is no longer valid.";
	}
    
	/**
	 * @see net.sourceforge.orbroker.QueryableConnection#getFinalizeWarning()
	 */
	protected String getFinalizeWarning() {
	    return "Executable was not released. " +
	    		"This can cause connection cross usage with unpredictable results.";
	}

	/**
	 * @inheritDoc
	 * @see net.sourceforge.orbroker.ExecutableConnection#markTransactionStarted()
	 */
	protected void markTransactionStarted() {
		if (this.transaction != null) {
			this.transaction.markTransactionStarted();
		}
	}

	protected boolean retryFailedStatement(SQLException e) {
		// Don't retry because it's an outside supplied connection.
    	return false;
    }

    /**
	 * @see net.sourceforge.orbroker.QueryableConnection#setConnectionProperties()
	 */
	protected void setConnectionProperties() throws SQLException {
	    // Do not touch connection.
		return;
	}
	
    /**
     * @see net.sourceforge.orbroker.BrokerConnection#shouldConnectionClose()
     */
    protected boolean shouldConnectionClose() {
    	// Connection handled externally, so never close.
        return false;
    }

    /**
     * Invalidate this Executable.
     */
    void release() {
        dropConnection();
    }
}