/*
 * Created on Feb 22, 2004
 */
package net.sourceforge.orbroker;

import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Savepoint;
import java.util.Collection;



/**
 * Executable connection.
 * 
 * @author Nils Kilden-Pedersen
 * @see net.sourceforge.orbroker.Broker#startTransaction()
 * @see net.sourceforge.orbroker.Broker#startTransaction(int)
 */
abstract class ExecutableConnection extends QueryableConnection {
	
	private static final int[] EMPTY_COUNT = new int[0];
	ExecutableConnection(Broker broker, Connection connection) {
	    super(broker, connection);
	}
	
	ExecutableConnection(Broker broker, Integer isolationLevel) throws BrokerException {
		super(broker, isolationLevel);
	}

	protected Statement getStatement(String statementID) throws BrokerException {
		markTransactionStarted();
		return super.getStatement(statementID);
	}
    
	protected abstract void markTransactionStarted();
	
	/**
	 * Execute statement. This is used for statements such as
	 * <ul>
	 * <li>UPDATE</li>
	 * <li>DELETE</li>
	 * <li>INSERT</li>
	 * <li>CALL</li>
	 * </ul>
	 * 
	 * @param statementID Statement id
	 * @return number of records affected.
	 * @throws BrokerException
	 * @throws ConstraintException
	 * @throws DeadlockException
	 */
	public int execute(String statementID) 
			throws BrokerException,
			ConstraintException,
			DeadlockException {
		Statement stm = getStatement(statementID);
		try {
			return stm.executeUpdate(getActiveConnection(), getConnectionContext());
		} catch (SQLException e) {
		    if (retryFailedStatement(e)) {
		        return execute(statementID);
		    }
			throwException(e);
		}
		// Not going to happen, but necessary for compiler
        return 0;
	}

	/**
	 * Execute a collection of parameters as a batch.
	 * 
	 * @param statementID Statement id
	 * @param batchParameterName
	 *                   The parameter name that will be repeated
	 * @param batchParameters
	 *                   The collection of batch parameter objects
	 * @return Total records affected, or a negative number if unknown
	 * @throws BrokerException
	 * @throws ConstraintException
	 * @throws DeadlockException
	 */
	public int executeBatch(String statementID, String batchParameterName,
			Collection batchParameters) throws BrokerException,
			ConstraintException, DeadlockException {

		int[] counts = executeBatch(statementID, batchParameterName,
				batchParameters.toArray());
		int totalCount = 0;
		for (int i = 0; i < counts.length; i++) {
			if (counts[i] < 0) {
				return counts[i];
			}
			totalCount += counts[i];
		}
		return totalCount;

	}

	/**
	 * Execute an array of parameters as a batch.
	 * 
	 * @param statementID Statement id
	 * @param batchParameterName
	 *                   The parameter name that will be repeated
	 * @param batchParameters
	 *                   The array of batch objects
	 * @return Total records affected per parameter, or a negative number if unknown
	 * @throws BrokerException
	 * @throws ConstraintException
	 * @throws DeadlockException
	 */
	public int[] executeBatch(String statementID, String batchParameterName,
			Object[] batchParameters) throws BrokerException,
			ConstraintException, DeadlockException {

		if (batchParameters == null || batchParameters.length == 0) {
			return EMPTY_COUNT;
		}
		
		if (!supportsBatch()) {
			throw new BrokerException(
					"Batch execution is not supported by driver");
		}

		int[] totalCount = EMPTY_COUNT;
		Statement stm = getStatement(statementID);
		try {
			totalCount = 
				stm.executeBatch(
			        getActiveConnection(), 
			        getConnectionContext(),
					batchParameterName, 
					batchParameters);
		} catch (SQLException e) {
		    if (retryFailedStatement(e)) {
		        return executeBatch(statementID, batchParameterName, batchParameters);
		    }
			throwException(e);
		}
        return totalCount;
	}

	/**
	 * Release savepoint.
	 * @param savepoint Savepoint
	 * @see Connection#releaseSavepoint(java.sql.Savepoint)
	 */
	public void releaseSavepoint(Savepoint savepoint) {
	    try {
            getActiveConnection().releaseSavepoint(savepoint);
        } catch (SQLException e) {
            throwException(e);
        }
	}

	/**
	 * Roll back to savepoint.
	 * @param savepoint Savepoint
	 * @see Connection#rollback(java.sql.Savepoint)
	 */
	public void rollback(Savepoint savepoint) {
	    try {
            getActiveConnection().rollback(savepoint);
        } catch (SQLException e) {
            throwException(e);
        }
	}
	
	/**
	 * Set savepoint.
	 * @return savepoint
	 * @see Connection#setSavepoint()
	 */
	public Savepoint setSavepoint() {
	    return setSavepoint(null);
	}
	
	/**
	 * Set named savepoint.
	 * @param name Savepoint name
	 * @return savepoint
	 * @see Connection#setSavepoint(String)
	 *	 */
	public Savepoint setSavepoint(String name) {
	    Savepoint sp = null;
	    try {
	        if (name != null) {
	            sp = getActiveConnection().setSavepoint(name);
	        } else {
	            sp = getActiveConnection().setSavepoint();
	        }
        } catch (SQLException e) {
            if (retryFailedStatement(e)) {
                return setSavepoint(name);
            }
            throwException(e);
        }
        return sp;
	}
}