package net.sourceforge.orbroker;

/*
 * Created on Feb 27, 2004
 */

import java.io.IOException;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.io.Reader;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.sql.Types;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.WeakHashMap;

import net.kildenpedersen.reflect.Reflector;
import net.kildenpedersen.util.ExpandingArrayList;

/**
 * JDBC type representation.
 * 
 * @author Nils Kilden-Pedersen
 */
final class JdbcType {
	private static class MappingStrategy {
		private final Constructor constructor;
		private final Object[] oneParm = new Object[1];
		private final Method resultRowMethod;
		private final Method staticMethod;
		
		private MappingStrategy(Method resultRowMethod) {
			this.resultRowMethod = resultRowMethod;
			this.constructor = null;
			this.staticMethod = null;
		}
		
		private MappingStrategy(Method resultRowMethod, Constructor constructor) {
			this.resultRowMethod = resultRowMethod;
			this.constructor = constructor;
			this.staticMethod = null;
		}
		
		private MappingStrategy(Method resultRowMethod, Method staticMethod) {
			this.resultRowMethod = resultRowMethod;
			this.staticMethod = staticMethod;
			this.constructor = null;
		}
		
		private Object getColumnValue(ResultRow row, int columnIndex) {
			this.oneParm[0] = invoke(this.resultRowMethod, row, columnIndex);

			if (this.constructor != null) {
				try {
					return this.constructor.newInstance(this.oneParm);
				} catch (Exception e) {
					throw new ReflectionException(e);
				}
			}

			if (this.staticMethod != null) {
				try {
					return this.staticMethod.invoke(null, this.oneParm);
				} catch (Exception e) {
					throw new ReflectionException(e);
				}
			}
			
			return this.oneParm[0];
			
		}
	}
	
    private static final Integer[] INTEGER_CACHE;
    private static final int INTEGER_OFFSET;
    private static final HashMap JDBC_TYPES = new HashMap();
    private static final HashMap JDBC_TYPES_BY_INT = new HashMap();
    private static final HashMap JDBC_TYPES_BY_PREFERRED_TYPE = new HashMap();
    private static final String JDBC_TYPES_FILE = "jdbc-types.csv";
	private static final WeakHashMap MAPPING_STRATEGIES = new WeakHashMap();
    private static final Class[] PARM_CLASS_INT = new Class[] { int.class };
    private static final Map PRIMITIVE_WRAPPERS = new HashMap();
    static {
        JDBC_TYPES.put("ARRAY", new Integer(Types.ARRAY));
        JDBC_TYPES.put("BIGINT", new Integer(Types.BIGINT));
        JDBC_TYPES.put("BINARY", new Integer(Types.BINARY));
        JDBC_TYPES.put("BIT", new Integer(Types.BIT));
        JDBC_TYPES.put("BLOB", new Integer(Types.BLOB));
        JDBC_TYPES.put("BOOLEAN", new Integer(Types.BOOLEAN));
        JDBC_TYPES.put("CHAR", new Integer(Types.CHAR));
        JDBC_TYPES.put("CLOB", new Integer(Types.CLOB));
        JDBC_TYPES.put("DATALINK", new Integer(Types.DATALINK));
        JDBC_TYPES.put("DATE", new Integer(Types.DATE));
        JDBC_TYPES.put("DECIMAL", new Integer(Types.DECIMAL));
        JDBC_TYPES.put("DISTINCT", new Integer(Types.DISTINCT));
        JDBC_TYPES.put("DOUBLE", new Integer(Types.DOUBLE));
        JDBC_TYPES.put("FLOAT", new Integer(Types.FLOAT));
        JDBC_TYPES.put("INTEGER", new Integer(Types.INTEGER));
        JDBC_TYPES.put("JAVA_OBJECT", new Integer(Types.JAVA_OBJECT));
        JDBC_TYPES.put("LONGVARBINARY", new Integer(Types.LONGVARBINARY));
        JDBC_TYPES.put("LONGVARCHAR", new Integer(Types.LONGVARCHAR));
        JDBC_TYPES.put("NUMERIC", new Integer(Types.NUMERIC));
        JDBC_TYPES.put("OTHER", new Integer(Types.OTHER));
        JDBC_TYPES.put("REAL", new Integer(Types.REAL));
        JDBC_TYPES.put("REF", new Integer(Types.REF));
        JDBC_TYPES.put("SMALLINT", new Integer(Types.SMALLINT));
        JDBC_TYPES.put("STRUCT", new Integer(Types.STRUCT));
        JDBC_TYPES.put("TIME", new Integer(Types.TIME));
        JDBC_TYPES.put("TIMESTAMP", new Integer(Types.TIMESTAMP));
        JDBC_TYPES.put("TINYINT", new Integer(Types.TINYINT));
        JDBC_TYPES.put("VARBINARY", new Integer(Types.VARBINARY));
        JDBC_TYPES.put("VARCHAR", new Integer(Types.VARCHAR));

        int lowest = Integer.MAX_VALUE;
        int highest = Integer.MIN_VALUE;
        Iterator types = JDBC_TYPES.values().iterator();
        while (types.hasNext()) {
            Integer type = (Integer) types.next();
            if (type.intValue() < lowest) {
                lowest = type.intValue();
            }
            if (type.intValue() > highest) {
                highest = type.intValue();
            }
        }

        INTEGER_OFFSET = lowest;
        INTEGER_CACHE = new Integer[highest - lowest];
        for (int i = 0; i < INTEGER_CACHE.length; i++) {
            INTEGER_CACHE[i] = new Integer(i + INTEGER_OFFSET);
        }

    }
    static {
        PRIMITIVE_WRAPPERS.put(int.class, Integer.class);
        PRIMITIVE_WRAPPERS.put(long.class, Long.class);
        PRIMITIVE_WRAPPERS.put(short.class, Short.class);
        PRIMITIVE_WRAPPERS.put(boolean.class, Boolean.class);
        PRIMITIVE_WRAPPERS.put(byte.class, Byte.class);
        PRIMITIVE_WRAPPERS.put(double.class, Double.class);
        PRIMITIVE_WRAPPERS.put(float.class, Float.class);
        PRIMITIVE_WRAPPERS.put(char.class, Character.class);
    }

    static {
        readJdbcTypes();
    }

    /**
     * Adds valid {@link ResultRow}methods to a given JDBC type. This method
     * should only be called from {@link #readJdbcTypes()}.
     * 
     * @param typeSeq
     * @param method
     */
    private static void addMethodsToJdbcTypes(Integer[] typeSeq,
            StringTokenizer method) {
        String methodName = null;
        int seqIdx = -1;
        while (method.hasMoreTokens()) {
            String token = method.nextToken();
            if (seqIdx == -1) {
                methodName = token;
            } else {
                if (!token.equals(".")) {
                	// Only valid types are 'x' and 'P'
                    JdbcType jdbcType = 
                    	(JdbcType) JDBC_TYPES_BY_INT.get(typeSeq[seqIdx]);
                    jdbcType.addGetter(methodName, token.equals("P"));
                }
            }
            seqIdx++;
        }
    }

    private static void cacheMappingStrategy(ResultRow row, int index, MappingStrategy strategy) {
		if (strategy == null) {
			return;
		}
		ExpandingArrayList strategyList = (ExpandingArrayList) MAPPING_STRATEGIES.get(row.getMappingStrategyKey());
		if (strategyList == null) {
			strategyList = new ExpandingArrayList(30);
			MAPPING_STRATEGIES.put(row.getMappingStrategyKey(), strategyList);
		}
		strategyList.set(index, strategy);
	}
    
    private static MappingStrategy getCachedMappingStrategy(ResultRow row, int index) {
		ExpandingArrayList strategyList = (ExpandingArrayList) MAPPING_STRATEGIES.get(row.getMappingStrategyKey());
		if (strategyList == null) {
			return null;
		}
		return (MappingStrategy) strategyList.get(index);
	}

    private static Object invoke(final Method method, final ResultRow row, final int columnIndex) {
        Integer integerColumn = getCachedInteger(columnIndex);
        Object[] parms = new Object[] {integerColumn};
        Object result;
        try {
            result = method.invoke(row, parms);
        } catch (InvocationTargetException e) {
        	throw new ReflectionException(e.getTargetException());
        } catch (Exception e) {
            throw new ReflectionException(e);
        }
        if (row.wasNull()) {
            return null;
        }
        return result;
    }

    /**
     * Build the jdbcTypesByName {@link java.util.Map}. This method relies on
     * the <code>jdbc-types.xsl</code> spreadsheet being exported as a .csv
     * file. Since Excel doesn't export empty cells as anything, underscores are
     * used for blanks. X's are used to indicate a valid mapping. The Excel
     * spreadsheet method is used for best visibility against the Sun
     * documentation. An XML structure would probably be better in terms of code
     * readability, but it's a short function, so spreadsheet clarity won.
     */
    private static void readJdbcTypes() {
        Reader isr = new InputStreamReader(
        		JdbcType.class.getResourceAsStream(JDBC_TYPES_FILE));
        LineNumberReader reader = new LineNumberReader(isr);
        String line;
        Integer[] typeSequence = null;
        try {
            while ((line = reader.readLine()) != null) {
                StringTokenizer tokens = new StringTokenizer(line, ",");
                if (reader.getLineNumber() == 1) {
                    tokens.nextElement(); // Drop method header
                    typeSequence = new Integer[tokens.countTokens()];
                    int i = 0;
                    while (tokens.hasMoreTokens()) {
                        String jdbcTypeName = tokens.nextToken();
                        JdbcType jdbcType = new JdbcType(jdbcTypeName);
                        Integer key = (Integer) JDBC_TYPES.get(jdbcTypeName);
                        JDBC_TYPES_BY_INT.put(key, jdbcType);
                        typeSequence[i++] = key;
                    }
                } else {
                    addMethodsToJdbcTypes(typeSequence, tokens);
                }
            }
        } catch (IOException e) {
            throw new ConfigurationException(e);
        } finally {
            try {
                reader.close();
                isr.close();
            } catch (IOException e) {
                // Whatever
            }
        }
    }

    static Integer getCachedInteger(int value) {
        return INTEGER_CACHE[value - INTEGER_OFFSET];
    }

    /**
     * Return JDBC type based on return type. Since there are several JDBC types
     * that have the same Java class as preferred class, the type returned is
     * arbitrary among those sharing the same preferred class.
     * 
     * @param returnType
     * @return JDBC type
     */
    static JdbcType getInstance(Class returnType) {
        return (JdbcType) JDBC_TYPES_BY_PREFERRED_TYPE.get(returnType);
    }

    static JdbcType getInstance(int type) {
        return (JdbcType) JDBC_TYPES_BY_INT.get(getCachedInteger(type));
    }

    private final String name;

    private Method preferredMethod;
    private Class preferredReturnType;
    private Class preferredReturnTypeWrapper = null;

    private final HashMap resultRowMethods = new HashMap();

    private JdbcType(String name) {
        this.name = name;
    }
    private void addGetter(String methodName, boolean preferred) {
        Method method = getResultRowMethod(methodName);
        Class returnType = method.getReturnType();
        Class primitiveWrapper = (Class) PRIMITIVE_WRAPPERS.get(returnType);
        if (preferred) {
            this.preferredMethod = method;
            this.preferredReturnType = returnType;
            if (!JDBC_TYPES_BY_PREFERRED_TYPE.containsKey(returnType)) {
                JDBC_TYPES_BY_PREFERRED_TYPE.put(returnType, this);
                if (primitiveWrapper != null) {
                    this.preferredReturnTypeWrapper = primitiveWrapper;
                    JDBC_TYPES_BY_PREFERRED_TYPE.put(primitiveWrapper, this);
                }
            }
            return;
        }

        // Make sure no other method returns the same class
        if (this.resultRowMethods.containsKey(returnType)) {
            Method clashMethod = (Method) this.resultRowMethods
                    .get(returnType);
            throw new ConfigurationException("For " + this.toString()
                    + ", method " + methodName + " and "
                    + clashMethod.getName() + " both return "
                    + Reflector.getClassName(returnType));
        }
        this.resultRowMethods.put(returnType, method);
        if (primitiveWrapper != null) {
            this.resultRowMethods.put(primitiveWrapper, method);
        }
    }
    /**
	 * @param row
	 * @param columnIndex
	 * @param requiredType
	 * @return mapping strategy
	 */
	private MappingStrategy determineMappingStrategy(final ResultRow row, int columnIndex, final Class requiredType) {

		MappingStrategy strategy = getCachedMappingStrategy(row, columnIndex);
		if (strategy != null) {
			return strategy;
		}

		Method method = findAlternativeResultRowMethod(requiredType);
        if (method != null) {
        	strategy = new MappingStrategy(method);
        	cacheMappingStrategy(row, columnIndex, strategy);
        	return strategy;
        }
        
        // Find constructor or static factory method strategy
        strategy = findConstructorStrategy(row, columnIndex, requiredType);
        if (strategy == null) {
        	strategy = findFactoryMethodStrategy(row, columnIndex, requiredType);
        }
        
		return strategy;
	}

    /**
	 * @param row
	 * @param columnIndex
	 * @param requiredType
	 * @return Mapping strategy, or null if none found
	 */
	private MappingStrategy findConstructorStrategy(ResultRow row, int columnIndex, Class requiredType) {
        Constructor[] constructors = requiredType.getConstructors();
        for (int i = 0; i < constructors.length; i++) {
            Class[] parmClasses = constructors[i].getParameterTypes();
            if (parmClasses.length == 1) {
                Class argClass = parmClasses[0];
                Method method = getPreferredResultRowMethod(argClass);
                if (method == null) {
                	method = findAlternativeResultRowMethod(argClass);
                }
                if (method != null) {
                	MappingStrategy strategy = new MappingStrategy(method, constructors[i]);
                    cacheMappingStrategy(row, columnIndex, strategy);
                    return strategy;
                }
            }
        }
        return null;
	}

	/**
	 * @param row
	 * @param columnIndex
	 * @param requiredType
	 * @return Mapping strategy, or null if none found
	 */
	private MappingStrategy findFactoryMethodStrategy(ResultRow row, int columnIndex, Class requiredType) {
	    Method[] methods = requiredType.getMethods();
	    for (int i = 0; i < methods.length; i++) {

	    	Method factoryMethod = methods[i];
	    	
	    	// Continue loop if not public static one-arg method 
	    	if (!Modifier.isStatic(factoryMethod.getModifiers())
	    			|| factoryMethod.getParameterTypes().length != 1
	    			|| !Modifier.isPublic(factoryMethod.getModifiers())) {
	    		continue;
	    	}

	    	// Continue loop if return type is not assignable to required type
	    	Class returnType = factoryMethod.getReturnType();
	    	if (requiredType != returnType 
	    			&& !requiredType.isAssignableFrom(returnType)) {
	    		continue;
	    	}

	    	Class argClass = factoryMethod.getParameterTypes()[0];
            Method method = getPreferredResultRowMethod(argClass);
            if (method == null) {
            	method = findAlternativeResultRowMethod(argClass);
            }
            if (method != null) {
            	MappingStrategy strategy = new MappingStrategy(method, factoryMethod);
                cacheMappingStrategy(row, columnIndex, strategy);
                return strategy;
            }
	    }
	    return null;
	}

	/**
     * Find alternative ResultRow method for required type. 
     * Only use this after trying {@link #getPreferredResultRowMethod(Class)}.
     * @param requiredType
     * @return method
     */
    private Method findAlternativeResultRowMethod(final Class requiredType) {
        Method method = (Method) this.resultRowMethods.get(requiredType);
        if (method != null) {
            return method;
        }
        // Try all return class as sub-classes of the required type.
        Iterator entries = this.resultRowMethods.entrySet().iterator();
        while (entries.hasNext()) {
            Map.Entry entry = (Map.Entry) entries.next();
            Class returnClass = (Class) entry.getKey();
            if (requiredType.isAssignableFrom(returnClass)) {
                return (Method) entry.getValue();
            }
			if (returnClass.isPrimitive()) {
				returnClass = (Class) PRIMITIVE_WRAPPERS.get(returnClass);
	            if (requiredType.isAssignableFrom(returnClass)) {
	                return (Method) entry.getValue();
	            }
			}
        }
        return null;
    }

    private Method getPreferredResultRowMethod(Class requiredType) {
        if (	requiredType == this.preferredReturnType ||
        		requiredType.isAssignableFrom(this.preferredReturnType)) {
        	return this.preferredMethod;
        }
        if (this.preferredReturnTypeWrapper != null) {
        	if (	requiredType == this.preferredReturnTypeWrapper ||
        			requiredType.isAssignableFrom(this.preferredReturnTypeWrapper)) {
        		return this.preferredMethod;
        	}
        }
        return null;
    }

    private Method getResultRowMethod(String methodName) {
        try {
            return ResultRow.class.getMethod(methodName, PARM_CLASS_INT);
        } catch (NoSuchMethodException e) {
            throw new ConfigurationException(e.toString());
        }
    }

    /*
    Object getColumnValue(ResultRow row, int column) {
        return getColumnValue(row, column, this.preferredReturnType);
    }
    */
    
    /**
     * @param row
     * @param columnIndex
     * @param requiredType
     * @param timeZone
     *                   Can be null, if none has been specified.
     * @return column value
     */
    Object getColumnValue(final ResultRow row, int columnIndex, final Class requiredType) {
        Method method = getPreferredResultRowMethod(requiredType);
        if (method != null) {
        	return invoke(method, row, columnIndex);
        }
        
        // Cannot use preferred method. Employ other strategies.
        MappingStrategy strategy = determineMappingStrategy(row, columnIndex, requiredType);

        if (strategy != null) {
        	return strategy.getColumnValue(row, columnIndex);
        }

        // If all else has failed, throw an exception
        String table = row.getTableName(columnIndex);
        String column = row.getColumnName(columnIndex);
        String columnType = row.getColumnTypeName(columnIndex);
        String msg = "Unable to map column " + column + " (" + columnType + ")"
                + " in table " + table + " to type '" + Reflector.getClassName(requiredType) + "'";
        throw new BrokerException(msg);

    }

	/**
	 * Get the JDBC name.
	 * @return Returns name.
	 */
	String getName() {
		return this.name;
	}

	/**
     * @see java.lang.Object#toString()
     */
    public String toString() {
        return "JDBC type: " + this.name;
    }

}