package net.sourceforge.orbroker;

import java.util.Collection;
import java.util.Iterator;

/*
 * Created on Apr 20, 2004
 *
 */

/**
 * @author Nils Kilden-Pedersen
*/
abstract class LazyCollection implements Collection {

	private final Broker broker;
	private final Collection collectionInstance;
	private final ConnectionContext context;
	private boolean awake = false;
	private final String statementID;
	
	protected LazyCollection(
			Broker broker, 
			String statementID, 
			ConnectionContext contextParms,
			Collection collection) {
		this.broker = broker;
		this.context = contextParms;
		this.statementID = statementID;
		this.collectionInstance = collection;
	}
	
	/**
	 * Return the collection instance after checking that it has
	 * been awake. Any {@link Collection} method implementations,
	 * or sub-interface implementations, <i>must</i>
	 * use this method for access.
	 * @return collection
	 */
	protected Collection getCollection() {
		wakeUp();
		return this.collectionInstance;
	}
	
	/**
	 * Make sure that the lazy collectionInstance is initialized.
	 */
	protected synchronized void wakeUp() {
		if (this.awake) {
			return;
		}
		
		Query qry = this.broker.startQuery();
		qry.addContext(this.context);
		try {
		    qry.selectMany(this.statementID, this.collectionInstance);
		} finally {
			/* 
			 * Use regular close() instead of internalClose(), because 
			 * lazy collections may be performed in a different thread
			 * than its owner.
			 */
		    qry.close(); 
		}
		this.awake = true;
	}

	/**
	 * @see java.util.Collection#add(java.lang.Object)
	 */
	public boolean add(Object object) {
		return getCollection().add(object);
	}

	/**
	 * @see java.util.Collection#addAll(java.util.Collection)
	 */
	public boolean addAll(Collection collection) {
		return getCollection().addAll(collection);
	}

	/**
	 * @see java.util.Collection#clear()
	 */
	public void clear() {
		getCollection().clear();
	}

	/**
	 * @see java.util.Collection#contains(java.lang.Object)
	 */
	public boolean contains(Object object) {
		return getCollection().contains(object);
	}

	/**
	 * @see java.util.Collection#containsAll(java.util.Collection)
	 */
	public boolean containsAll(Collection collection) {
		return getCollection().containsAll(collection);
	}

	/**
	 * @see java.util.Collection#isEmpty()
	 */
	public boolean isEmpty() {
		return getCollection().isEmpty();
	}

	/**
	 * @see java.util.Collection#iterator()
	 */
	public Iterator iterator() {
		return getCollection().iterator();
	}

	/**
	 * @see java.util.Collection#remove(java.lang.Object)
	 */
	public boolean remove(Object object) {
		return getCollection().remove(object);
	}

	/**
	 * @see java.util.Collection#removeAll(java.util.Collection)
	 */
	public boolean removeAll(Collection collection) {
		return getCollection().removeAll(collection);
	}

	/**
	 * @see java.util.Collection#retainAll(java.util.Collection)
	 */
	public boolean retainAll(Collection collection) {
		return getCollection().retainAll(collection);
	}
	
	/**
	 * @see java.util.Collection#size()
	 */
	public int size() {
		return getCollection().size();
	}

	/**
	 * @see java.util.Collection#toArray()
	 */
	public Object[] toArray() {
		return getCollection().toArray();
	}

	/**
	 * @see java.util.Collection#toArray(java.lang.Object[])
	 */
	public Object[] toArray(Object[] array) {
		return getCollection().toArray(array);
	}

}
