/*
 * Created on Feb 22, 2004
 */
package net.sourceforge.orbroker;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;



/**
 * An abstract queryable connection.
 * 
 * @author Nils Kilden-Pedersen
 * @see net.sourceforge.orbroker.Query
 * @see net.sourceforge.orbroker.Transaction
 */

abstract class QueryableConnection extends BrokerConnection {
	
	/**
	 * Check that query only returned one row as expected.
	 * @param stmId Statement id
	 * @param qr Query result
	 * @throws MoreThanOneRowException
	 */
	private static final void checkForOne(String stmId, QueryResult qr) throws MoreThanOneRowException {
		if (qr.nextRow()) {
			String msg = "Statement '" + stmId + "' returned more than 1 row.";
			throw new MoreThanOneRowException(msg);
		}
	}

	private static final void checkRowRange(int startRow, int rowCount) {
		if (startRow < 1) {
			throw new IllegalArgumentException(
					"Starting row cannot be less than 1. Passed as " + startRow);
		}
		if (rowCount < 1) {
			throw new IllegalArgumentException(
					"Row count must be greater than 0. Passed as " + rowCount);
		}
	}

	protected QueryableConnection(Broker broker, Integer isolationLevel) {
		super(broker, isolationLevel);
	}

	protected QueryableConnection(Broker broker, Connection connection) {
		super(broker, connection);
	}

	/**
	 * @param stm
	 * @param useScrollable
	 * @return query result
	 * @throws BrokerException
	 */
	protected final QueryResult executeQuery(Statement stm, boolean useScrollable) throws BrokerException {
		try {
			return stm.executeQuery(
				getActiveConnection(),
				useScrollable,
				getConnectionContext());
		} catch (SQLException e) {
		    if (retryFailedStatement(e)) {
		        return executeQuery(stm, useScrollable);
		    } 
		    throw new QueryException(e);
		}
	}
	
	/**
	 * Return a list of 0..n objects. An empty {@link List} will be returned if
	 * result set is empty.
	 * 
	 * @param statementID Statement id
	 * @return List of result objects
	 * @throws BrokerException
	 */
	public final List selectMany(String statementID) throws BrokerException {

		List resultList = new ArrayList();
		selectMany(statementID, resultList);
		return resultList;
	}

	/**
	 * Fill Collection with 0..n objects. The supplied Collection does not have
	 * to be empty. Objects are added to Collection.
	 * 
	 * @param statementID Statement id
	 * @param resultCollection
	 *                The collection where the result will be added.
	 * @return Number of objects added to collection.
	 * @throws BrokerException
	 */
	public final int selectMany(String statementID, Collection resultCollection) throws BrokerException {

		int initCollectionSize = resultCollection.size();
		Statement stm = getStatement(statementID);
		QueryResult qr = executeQuery(stm, false);
		ResultRow row = qr.getResultRow();
		try {
			while (qr.nextRow()) {
				Object obj = stm.buildResultObject(row);
				resultCollection.add(obj);
			}
		} finally {
		    qr.close();
		}
		return resultCollection.size() - initCollectionSize;
	}

	/**
	 * Return a live iterator. Very useful for large rows
	 * that, for memory concerns, need to be built a few at a time, 
	 * controlled by the fetch size. The Iterator must be iterated
	 * with the source object active, i.e. a 
	 * {@link Query} or {@link Transaction} must be open.
	 * @param statementID Statement id
	 * @param fetchSize Row buffer size
	 * @return Streamed {@link Iterator} of result objects.
	 */
	public Iterator iterate(String statementID, int fetchSize) {
	    Statement stm = getStatement(statementID);
	    QueryResult qr = executeQuery(stm, false);
	    qr.setFetchSize(fetchSize);
	    return new ResultSetIterator(qr, stm);
	}

	/**
	 * Close iterator. Use this to close a live iterator returned by
	 *  {@link #iterate(String, int)}. Only necessary if iterator has
	 *  not been exhausted by {@link Iterator#hasNext() hasNext()}
	 *  returning <code>false</code>
	 * @param iterator The not fully iterated iterator.
	 * @see #iterate(String, int)
	 */
	public final void closeIterator(Iterator iterator) {
		if (iterator instanceof ResultSetIterator) {
			((ResultSetIterator) iterator).closeIterator();
		}
	}
	
	/**
	 * Fill Collection with 0..n objects. The supplied Collection does not have
	 * to be empty. Objects are added to Collection.
	 * 
	 * @param statementID Statement id
	 * @param resultCollection
	 *                The collection where the result will be added.
	 * @param startRow The start row to return. First row is 1.
	 * @param rowCount The number of rows to return
	 * @return Number of objects added to collection.
	 * @throws BrokerException
	 */
	public final int selectMany(
			String statementID, 
			Collection resultCollection,
			int startRow,
			int rowCount) throws BrokerException {
	
		checkRowRange(startRow, rowCount);
		int initCollectionSize = resultCollection.size();
		Statement stm = getStatement(statementID);
		int rowCounter = 0;
		QueryResult qr = executeQuery(stm, supportsScrollable());
		qr.setFetchSize(startRow + rowCount -1);
		ResultRow row = qr.getResultRow(startRow, supportsScrollable());
		qr.setFetchSize(rowCount);
		try {
			while (rowCounter < rowCount && qr.nextRow()) {
				Object obj = stm.buildResultObject(row);
				resultCollection.add(obj);
				rowCounter++;
			}
		} finally {
		    qr.close();
		}
		return resultCollection.size() - initCollectionSize;
	}
	
	/**
	 * Return a list of 0..n objects. An empty {@link List} will be returned if
	 * result set is empty or if <code>startRow</code> is greater than 
	 * the result set size. 
	 * 
	 * @param statementID Statement id
	 * @param startRow The start row to return. First row is 1.
	 * @param rowCount The number of rows to return
	 * @return List of result objects
	 * @throws BrokerException
	 */
	public final List selectMany(
			String statementID, 
			int startRow, int rowCount) 
			throws BrokerException {
		
		List resultList = new ArrayList(rowCount);
		selectMany(statementID, resultList, startRow, rowCount);
		return resultList;
	}

	/**
	 * Return a single result object. 
	 * 
	 * @param statementID Statement id
	 * @return the result object, or <code>null</code> if query returned 0
	 *            records.
	 * @throws BrokerException
	 * @throws MoreThanOneRowException
	 */
	public final Object selectOne(String statementID) throws BrokerException, MoreThanOneRowException {
		Statement stm = getStatement(statementID);
		QueryResult qr = executeQuery(stm, false);
		qr.setFetchSize(1);
		Object obj;
		ResultRow row = qr.getResultRow();
		try {
			if (qr.nextRow()) {
				obj = stm.buildResultObject(row);
				checkForOne(statementID, qr);
			} else {
				obj = null;
			}
		} finally {
		    qr.close();
		}
		return obj;
	}

	/**
	 * Use supplied result object to apply values to. Any constructor or 
	 * factory method defined for mapping will be ignored.  
	 * @param statementID Statement id
	 * @param resultObject result object to map to
	 * @return <code>false</code> if query didn't return a record.
	 * @throws BrokerException
	 * @throws MoreThanOneRowException
	 */
	public final boolean selectOne(String statementID, Object resultObject) 
			throws BrokerException, 
			MoreThanOneRowException {
	    
		Statement stm = getStatement(statementID); 
		QueryResult qr = executeQuery(stm, false);
		qr.setFetchSize(1);
		ResultRow row = qr.getResultRow();
		try {
			if (!qr.nextRow()) {
				return false;
			}
			stm.buildResultObject(row, resultObject);
			checkForOne(statementID, qr);
		} finally {
		    qr.close();
		}
		return true;
	}
	
	/**
	 * Return one result object from a query containing many.
	 * Returns <code>null</code> if query is empty or returnRow is out of range.
	 * @param statementID Statement id
	 * @param fromRow Row to return. First row is 1.
	 * @return result object from row
	 * @throws BrokerException
	 */
	public Object selectOneFromMany(String statementID, int fromRow) throws BrokerException {
		int rowCount = 1;
		checkRowRange(fromRow, rowCount);
	    Statement stm = getStatement(statementID);
		QueryResult qr = executeQuery(stm, supportsScrollable());
		qr.setFetchSize(fromRow);
		ResultRow row = qr.getResultRow(fromRow, supportsScrollable());
		qr.setFetchSize(rowCount);
		Object obj;
		try {
			if (qr.nextRow()) {
				obj = stm.buildResultObject(row);
			} else {
				obj = null;
			}
		} finally {
		    qr.close();
		}
		return obj;
	}
	

}