package net.sourceforge.orbroker;

import java.lang.reflect.Modifier;
import java.util.HashMap;
import java.util.Map;

/*
 * Created on Mar 9, 2004
*/

/**
 * @author Nils Kilden-Pedersen
 */
final class ResultObjectDefinition {
	private final Broker broker;
    private Map delegateToResultObjects;
    private Column discriminatorColumn;
	private final String extendsId;
	private final String id; 
	private ResultObjectInstantiator instantiator;
	
	private ResultKeyDef keyDefinition;
	private final SetterList setterList = new SetterList();
    
    private final Class type;
	
	ResultObjectDefinition(String id, Class type, String extendsId, ResultKeyDef keyDef, Broker broker) {
	    this.type = type;
		this.id = id;
		if (isInstantiable() && keyDef == null && extendsId == null) {
		    String msg = "No key columns are defined for '" + id +"'";
		    throw new ConfigurationException(msg);
		}
		this.extendsId = extendsId;
		this.keyDefinition = keyDef;
		this.broker = broker;
	}

	void addDynamicResultObject(
			ResultObjectDefinition resultObjectDef, 
			String onValue) {
		assert this.discriminatorColumn != null;
		this.delegateToResultObjects.put(onValue, resultObjectDef);
	}

	void addSetter(Setter setter) {
		this.setterList.add(setter);
	}
	
	/**
	 * Builds the result object.
	 * @param row
	 * @return result object
	 * @see net.sourceforge.orbroker.ValueType#getValue(net.sourceforge.orbroker.ResultRow)
	 */
	Object buildObject(ResultRow row)  {
		assert this.instantiator != null:
			"Constructor or factory method has not been defined for '" 
			+ this.id + "'.";

		ResultKey key = this.keyDefinition.getResultKey(row);
		if (key == null) {
		    /*
		     * Key columns can be null if the columns
		     * are from an OUTER JOIN, so return null, since
		     * object cannot be created.
		     */
		    return null;
		}
		Object resultObject = null;
		if (key.isValidKey()) {
			resultObject = row.getCachedResultObject(key);
		}
		if (resultObject == null) {
			resultObject = this.instantiator.newInstance(row);
			row.cacheResultObject(key, resultObject);
			setValues(resultObject, row);
		}
		return resultObject;
	}
	
    /**
	 * Close object for further modification.
	 */
	void close() {
	    assert 
	    	(isInstantiable() && (this.keyDefinition != null || this.extendsId != null)) ||
	    	!isInstantiable():
	        "Result object '" + getId() + "' has no key definition";
	    assert 
	    	(isInstantiable() && hasInstantiator()) ||
	    	!isInstantiable():
	        "Result object '" + getId() + "' has no constructor";
	    	
		this.setterList.close();
	}
	
	Broker getBroker() {
		return this.broker;
	}
	/**
	 * @return Returns the extendsId.
	 */
	String getExtendsId() {
		return this.extendsId;
	}

	/**
	 * @return Returns the id.
	 */
	String getId() {
		return this.id;
	}
	
	ResultObjectDefinition getRuntimeResultObjectDef(ResultRow row) throws ConfigurationException {
		if (this.discriminatorColumn == null || row == null) {
			return this;
		}
		String columnValue = (String) this.discriminatorColumn.getValue(row);
		ResultObjectDefinition dynamicFactory =
			(ResultObjectDefinition)
				this.delegateToResultObjects.get(columnValue);
		if (dynamicFactory == null) {
			return this;
		}
		return dynamicFactory;
	}
    /**
     * @return Returns type.
     */
    Class getType() {
        return this.type;
    }
	
	boolean hasInstantiator() {
		return this.instantiator != null;
	}
	
	/**
	 * Initialize this builder with the super builder it extends.
	 * @param parentFactory
	 */
	void initializeWith(ResultObjectDefinition parentFactory) {
		assert parentFactory != null 
			: "Result object to extend from is null.";
		assert !parentFactory.id.equals(this.id)
			: "Result object has same id as parent.";
		
		if (parentFactory.hasInstantiator()) {
		    this.instantiator = parentFactory.instantiator.copy(this.type);
		}
		this.keyDefinition = parentFactory.keyDefinition;
		SetterList superSetters = parentFactory.setterList;
		for (int i = 0; i < superSetters.size(); i++) {
			this.addSetter(superSetters.getSetter(i));
		}
	}
    
    boolean isClosed() {
    	return this.setterList.isClosed();
    }

    /**
     * Is type instantiable? That means it cannot be interface or abstract class.
     * @return true, if instantiable
     */
    boolean isInstantiable() {
        return !Modifier.isAbstract(this.type.getModifiers());
    }

	/**
	 * @param discriminatorColumn The discriminatorColumn to set.
	 */
	void setDiscriminatorColumn(Column discriminatorColumn) {
		assert this.discriminatorColumn == null;
		this.discriminatorColumn = discriminatorColumn;
		this.delegateToResultObjects = new HashMap();
	}

	/**
	 * @param instantiator  The instantiator to set.
	 */
	void setInstantiator(ResultObjectInstantiator instantiator) {
		this.instantiator = instantiator;
	}

	/**
	 * Set the values on an already created result object
	 * @param resultObject
	 * @param row
	 */
	void setValues(Object resultObject, ResultRow row) {
		for (int i = 0; i < this.setterList.size(); i++) {
			this.setterList.getSetter(i).
				setValue(resultObject, row);
		}
	}
	
	/**
	 * @inheritDoc
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	public boolean equals(Object obj) {
		if (ResultObjectDefinition.class == obj.getClass()) {
			ResultObjectDefinition comp = (ResultObjectDefinition) obj; 
			return 
				this.id.equals(comp.id) &&
				this.broker.equals(comp.broker);
		}
		return false;
	}
	
	/**
	 * @inheritDoc
	 * @see java.lang.Object#hashCode()
	 */
	public int hashCode() {
		return this.id.hashCode();
	}
	
	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "Result object: " + this.id;
	}
	
}
