/*
 * Created on Mar 3, 2004
 */
package net.sourceforge.orbroker;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.math.BigDecimal;
import java.net.URL;
import java.sql.Array;
import java.sql.Blob;
import java.sql.Clob;
import java.sql.Connection;
import java.sql.Date;
import java.sql.Ref;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Time;
import java.sql.Timestamp;
import java.sql.Types;
import java.util.ArrayList;
import java.util.Map;

/**
 * Representation of a row. This wraps a result set with most ResultSet methods
 * available. The next() method is not available to prevent unauthorized
 * scrolling.
 * 
 * @author Nils Kilden-Pedersen
 */
final class ResultRow {

	/**
	 * @param metaData
	 * @return array of readable columns
	 * @throws SQLException
	 */
	private static int[] buildReadableColumns(ResultSetMetaData metaData) {
		int[] columns;
		try {
			columns = new int[metaData.getColumnCount()];
		} catch (SQLException e) {
			throw new BrokerException(e);
		}
		for (int i = 0; i < columns.length; i++) {
			columns[i] = i + 1;
		}
		return columns;
	}

	private static void checkLength(long length, int columnType,
			String desiredType) {
		if (length > Integer.MAX_VALUE) {
			String msg = JdbcType.getInstance(columnType).getName()
					+ " length of " + length
					+ " exceeds maximum size for conversion to " + desiredType;
			throw new BrokerException(msg);
		}
	}

	private ArrayList aliasStack = null;

	private final Map cachedResultObjects;

	private final Connection connection;

	private final ConnectionContext context;

	private Map currentAliases = null;

	private final ResultSetMetaData metaData;

	private final int[] readableColumns;

	private final ResultSet resultSet;

	private final Statement statement;

	/**
	 * Constructor.
	 * 
	 * @param resultSet
	 * @param context
	 * @param connection
	 */
	ResultRow(ResultSet resultSet, ConnectionContext context,
			Connection connection) {
		this.context = context;
		this.resultSet = resultSet;

		this.cachedResultObjects = Broker.getCachedObjects(connection);
		this.connection = connection;
		try {
			this.metaData = resultSet.getMetaData();
		} catch (SQLException e) {
			throw new BrokerException(e);
		}
		if (this.metaData == null) {
			throw new BrokerException("JDBC driver does not return metadata.");
		}
		if (resultSet.getClass() == ResultParameters.class) {
			ResultParameters resultParms = (ResultParameters) resultSet;
			this.readableColumns = resultParms.getOutputParameters();
		} else {
			this.readableColumns = buildReadableColumns(this.metaData);
		}

		Statement tempStatement;
		try {
			tempStatement = resultSet.getStatement();
		} catch (SQLException e) {
			tempStatement = null;
		}
		this.statement = tempStatement;

	}

	/**
	 * Get column JDBC type. Guaranteed not to return null.
	 * 
	 * @param columnIndex
	 * @param requiredType
	 * @return JDBC type
	 */
	private JdbcType getColumnJdbcType(int columnIndex, Class requiredType) {
		JdbcType jdbcType = JdbcType.getInstance(getColumnType(columnIndex));
		if (jdbcType == null) {
			jdbcType = JdbcType.getInstance(requiredType);
		}
		if (jdbcType != null) {
			return jdbcType;
		}
		String msg = "Unknown JDBC type: " + getColumnTypeName(columnIndex)
				+ ". Cannot find suitable type for " + requiredType.toString();
		throw new ConfigurationException(msg);
	}

	private int getColumnType(int columnIndex) {
		try {
			return this.metaData.getColumnType(columnIndex);
		} catch (SQLException e) {
			throw new BrokerException(e);
		}
	}

	void cacheResultObject(ResultKey pk, Object resultObject) {
		if (pk.isValidKey()) {
			this.cachedResultObjects.put(pk, resultObject);
		}
	}

	int findColumn(String columnName) {
		if (this.currentAliases == null && this.aliasStack != null
				&& !this.aliasStack.isEmpty()) {
			this.currentAliases = (Map) this.aliasStack.get(this.aliasStack
					.size() - 1);
		}
		if (this.currentAliases != null) {
			String queryColumn = (String) this.currentAliases.get(columnName);
			if (queryColumn != null) {
				columnName = queryColumn;
			}
		}

		try {
			return this.resultSet.findColumn(columnName);
		} catch (SQLException e) {
			String msg = "Cannot find column '" + columnName + "'";
			try {
				String columnList = " among available: ";
				int columns = this.metaData.getColumnCount();
				for (int column = 1; column <= columns; column++) {
					if (column > 1) {
						columnList += ", ";
					}
					columnList += this.metaData.getColumnName(column);
				}
				msg += columnList;
			} catch (SQLException e2) {
				// Ok, too bad
			}
			throw new BrokerException(msg, e);
		}
	}

	Object getCachedResultObject(ResultKey pk) {
		if (pk.isValidKey()) {
			return this.cachedResultObjects.get(pk);
		}
		return null;
	}

	/**
	 * @param columnIndex
	 * @return column label
	 * @see ResultSetMetaData#getColumnName(int)
	 */
	String getColumnLabel(int columnIndex) {
		try {
			return this.metaData.getColumnLabel(columnIndex);
		} catch (SQLException e) {
			throw new BrokerException(e);
		}
	}

	/**
	 * @param columnIndex
	 * @return column name
	 * @see ResultSetMetaData#getColumnName(int)
	 */
	String getColumnName(int columnIndex) {
		try {
			return this.metaData.getColumnName(columnIndex);
		} catch (SQLException e) {
			return "[unknown column]";
		}
	}

	/**
	 * @param columnIndex
	 * @return column type name
	 * @see ResultSetMetaData#getColumnTypeName(int)
	 */
	String getColumnTypeName(int columnIndex) {
		try {
			return this.metaData.getColumnTypeName(columnIndex);
		} catch (SQLException e) {
			throw new BrokerException(e);
		}
	}

	Object getColumnValue(int columnIndex, Class requiredType) {
		return getColumnJdbcType(columnIndex, requiredType).getColumnValue(
				this, columnIndex, requiredType);
	}

	Connection getConnection() {
		return this.connection;
	}

	Object getMappingStrategyKey() {
		if (this.statement != null) {
			return this.statement;
		}
		return this.metaData;
	}

	int[] getReadableColumns() {
		return this.readableColumns;
	}

	/**
	 * @param columnIndex
	 * @return table name
	 * @see ResultSetMetaData#getTableName(int)
	 */
	String getTableName(int columnIndex) {
		String unknownTable = "[unknown table]";
		try {
			String tableName = this.metaData.getTableName(columnIndex);
			if (tableName != null && tableName.trim().length() > 0) {
				return tableName;
			}
		} catch (SQLException e) {
			return unknownTable;
		}
		return unknownTable;
	}

	TextReplacements getTextReplacements() {
		return this.context.getTextReplacements();
	}

	/**
	 * @return Type
	 * @see java.sql.ResultSet#getType()
	 */
	int getType() {
		try {
			return this.resultSet.getType();
		} catch (SQLException e) {
			throw new BrokerException(e);
		}
	}

	/**
	 * Pulls the current map of aliases off the stack.
	 * 
	 * @see #pushAliases(Map)
	 */
	void pullAliases() {
		Object removedMap = this.aliasStack.remove(this.aliasStack.size() - 1);
		assert this.currentAliases == removedMap : "Current alias map does not match stack.";
		this.currentAliases = null;
	}

	/**
	 * Pushes a map of aliases on the alias stack.
	 * 
	 * @param aliases
	 * @see #pullAliases()
	 */
	void pushAliases(Map aliases) {
		if (this.aliasStack == null) {
			this.aliasStack = new ArrayList(15);
		}
		this.aliasStack.add(aliases);
	}

	/**
	 * @return true, if column was null.
	 * @see java.sql.ResultSet#wasNull()
	 */
	boolean wasNull() {
		try {
			return this.resultSet.wasNull();
		} catch (SQLException e) {
			throw new BrokerException(e);
		}
	}

	/**
	 * @param columnIndex
	 * @return Array
	 * @see java.sql.ResultSet#getArray(int)
	 */
	public Array getArray(int columnIndex) {
		try {
			return this.resultSet.getArray(columnIndex);
		} catch (SQLException e) {
			throw new BrokerException(e);
		}
	}

	/**
	 * @param columnIndex
	 * @return InputStream
	 * @see java.sql.ResultSet#getAsciiStream(int)
	 */
	public InputStream getAsciiStream(int columnIndex) {
		int columnType = getColumnType(columnIndex);
		try {
			if (columnType == Types.CLOB) {
				Clob clob = this.resultSet.getClob(columnIndex);
				if (clob == null)
					return null;
				return clob.getAsciiStream();
			}
			return this.resultSet.getAsciiStream(columnIndex);
		} catch (SQLException e) {
			throw new BrokerException(e);
		}
	}

	/**
	 * @param columnIndex
	 * @return BigDecimal
	 * @see java.sql.ResultSet#getBigDecimal(int)
	 */
	public BigDecimal getBigDecimal(int columnIndex) {
		try {
			return this.resultSet.getBigDecimal(columnIndex);
		} catch (SQLException e) {
			throw new BrokerException(e);
		}
	}

	/**
	 * @param columnIndex
	 * @return InputStream
	 * @see java.sql.ResultSet#getBinaryStream(int)
	 */
	public InputStream getBinaryStream(int columnIndex) {
		int columnType = getColumnType(columnIndex);
		try {
			if (columnType == Types.BLOB) {
				Blob blob = this.resultSet.getBlob(columnIndex);
				if (blob == null)
					return null;
				return blob.getBinaryStream();
			}
			return this.resultSet.getBinaryStream(columnIndex);

		} catch (SQLException e) {
			throw new BrokerException(e);
		}
	}

	/**
	 * @param columnIndex
	 * @return Blob
	 * @see java.sql.ResultSet#getBlob(int)
	 */
	public Blob getBlob(int columnIndex) {
		try {
			return this.resultSet.getBlob(columnIndex);
		} catch (SQLException e) {
			throw new BrokerException(e);
		}
	}

	/**
	 * @param columnIndex
	 * @return boolean
	 * @see java.sql.ResultSet#getBoolean(int)
	 */
	public boolean getBoolean(int columnIndex) {
		try {
			return this.resultSet.getBoolean(columnIndex);
		} catch (SQLException e) {
			throw new BrokerException(e);
		}
	}

	/**
	 * @param columnIndex
	 * @return byte
	 * @see java.sql.ResultSet#getByte(int)
	 */
	public byte getByte(int columnIndex) {
		try {
			return this.resultSet.getByte(columnIndex);
		} catch (SQLException e) {
			throw new BrokerException(e);
		}
	}

	/**
	 * @param columnIndex
	 * @return byte[]
	 * @see java.sql.ResultSet#getBytes(int)
	 */
	public byte[] getBytes(int columnIndex) {
		int columnType = getColumnType(columnIndex);
		try {
			switch (columnType) {

			case Types.BLOB:
				Blob blob = this.resultSet.getBlob(columnIndex);
				if (blob == null)
					return null;
				checkLength(blob.length(), columnType, "byte[]");
				return blob.getBytes(1, (int) blob.length());

			case Types.CLOB:
				Clob clob = this.resultSet.getClob(columnIndex);
				if (clob == null)
					return null;
				checkLength(clob.length(), columnType, "byte[]");
				byte[] bytes = new byte[(int) clob.length()];
				BufferedInputStream is = new BufferedInputStream(clob
						.getAsciiStream());
				try {
					is.read(bytes);
				} catch (IOException e) {
					throw new BrokerException(e);
				}
				return bytes;

			default:
				return this.resultSet.getBytes(columnIndex);

			}
		} catch (SQLException e) {
			throw new BrokerException(e);
		}
	}

	/**
	 * @param columnIndex
	 * @return Reader
	 * @see java.sql.ResultSet#getCharacterStream(int)
	 */
	public Reader getCharacterStream(int columnIndex) {
		int columnType = getColumnType(columnIndex);
		try {
			switch (columnType) {

			case Types.CLOB:
				Clob clob = this.resultSet.getClob(columnIndex);
				if (clob == null)
					return null;
				return clob.getCharacterStream();

			case Types.BLOB:
				Blob blob = this.resultSet.getBlob(columnIndex);
				if (blob == null)
					return null;
				return new InputStreamReader(blob.getBinaryStream());

			default:
				return this.resultSet.getCharacterStream(columnIndex);
			}
		} catch (SQLException e) {
			throw new BrokerException(e);
		}
	}

	/**
	 * @param columnIndex
	 * @return Clob
	 * @see java.sql.ResultSet#getClob(int)
	 */
	public Clob getClob(int columnIndex) {
		try {
			return this.resultSet.getClob(columnIndex);
		} catch (SQLException e) {
			throw new BrokerException(e);
		}
	}

	/**
	 * @param columnIndex
	 * @return Date
	 * @see java.sql.ResultSet#getDate(int)
	 */
	public Date getDate(int columnIndex) {
		try {
			return this.resultSet.getDate(columnIndex);
		} catch (SQLException e) {
			throw new BrokerException(e);
		}
	}

	/**
	 * @param columnIndex
	 * @return double
	 * @see java.sql.ResultSet#getDouble(int)
	 */
	public double getDouble(int columnIndex) {
		try {
			return this.resultSet.getDouble(columnIndex);
		} catch (SQLException e) {
			throw new BrokerException(e);
		}
	}

	/**
	 * @param columnIndex
	 * @return float
	 * @see java.sql.ResultSet#getFloat(int)
	 */
	public float getFloat(int columnIndex) {
		try {
			return this.resultSet.getFloat(columnIndex);
		} catch (SQLException e) {
			throw new BrokerException(e);
		}
	}

	/**
	 * @param columnIndex
	 * @return int
	 * @see java.sql.ResultSet#getInt(int)
	 */
	public int getInt(int columnIndex) {
		try {
			return this.resultSet.getInt(columnIndex);
		} catch (SQLException e) {
			throw new BrokerException(e);
		}
	}

	/**
	 * @param columnIndex
	 * @return long
	 * @see java.sql.ResultSet#getLong(int)
	 */
	public long getLong(int columnIndex) {
		try {
			return this.resultSet.getLong(columnIndex);
		} catch (SQLException e) {
			throw new BrokerException(e);
		}
	}

	/**
	 * @param columnIndex
	 * @return Object
	 * @see java.sql.ResultSet#getObject(int)
	 */
	public Object getObject(int columnIndex) {
		int columnType = getColumnType(columnIndex);
		try {
			// Only boolean is not supported as getObject by JDBC.
			if (columnType == Types.BOOLEAN) {
				boolean value = this.resultSet.getBoolean(columnIndex);
				return Boolean.valueOf(value);
			}
			return this.resultSet.getObject(columnIndex);
		} catch (SQLException e) {
			throw new BrokerException(e);
		}
	}

	/**
	 * @param columnIndex
	 * @return Ref
	 * @see java.sql.ResultSet#getRef(int)
	 */
	public Ref getRef(int columnIndex) {
		try {
			return this.resultSet.getRef(columnIndex);
		} catch (SQLException e) {
			throw new BrokerException(e);
		}
	}

	/**
	 * @param columnIndex
	 * @return short
	 * @see java.sql.ResultSet#getShort(int)
	 */
	public short getShort(int columnIndex) {
		try {
			return this.resultSet.getShort(columnIndex);
		} catch (SQLException e) {
			throw new BrokerException(e);
		}
	}

	/**
	 * @param columnIndex
	 * @return String
	 * @see java.sql.ResultSet#getString(int)
	 */
	public String getString(int columnIndex) {
		int columnType = getColumnType(columnIndex);
		try {
			switch (columnType) {

			case Types.CLOB:
				Clob clob = this.resultSet.getClob(columnIndex);
				if (clob == null)
					return null;
				checkLength(clob.length(), columnType, "String");
				return clob.getSubString(1, (int) clob.length());

			case Types.BLOB:
				byte[] bytes = getBytes(columnIndex);
				if (bytes == null)
					return null;
				return new String(bytes);

			default:
				return this.resultSet.getString(columnIndex);

			}
		} catch (SQLException e) {
			throw new BrokerException(e);
		}
	}

	/**
	 * @param columnIndex
	 * @return Time
	 * @see java.sql.ResultSet#getTime(int)
	 */
	public Time getTime(int columnIndex) {
		try {
			return this.resultSet.getTime(columnIndex);
		} catch (SQLException e) {
			throw new BrokerException(e);
		}
	}

	/**
	 * @param columnIndex
	 * @return Timestamp
	 * @see java.sql.ResultSet#getTimestamp(int)
	 */
	public Timestamp getTimestamp(int columnIndex) {
		try {
			return this.resultSet.getTimestamp(columnIndex);
		} catch (SQLException e) {
			throw new BrokerException(e);
		}
	}

	/**
	 * @param columnIndex
	 * @return URL
	 * @see java.sql.ResultSet#getURL(int)
	 */
	public URL getURL(int columnIndex) {
		try {
			return this.resultSet.getURL(columnIndex);
		} catch (SQLException e) {
			throw new BrokerException(e);
		}
	}
}