/*
 * Created on Feb 25, 2005
 */
package net.sourceforge.orbroker;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;

/**
 * This class represents an immutable parsed statement.
 * @author Nils Kilden-Pedersen
 */
final class SQLParser {
	private static final String[] BREAK_CHARS = new String[]{
	" ", ",", "(", ")", "=", ">", "<", "|",
	"-", "+", "*", "/", "'", ";", "\"", "^",
	"%", "$", "#", "{", "}"};
	/**
	 * Is parameter name valid?
	 * @param parameterName
	 * @throws ConfigurationException
	 */
	private static void checkParameterName(String parameterName) throws ConfigurationException {
		if (parameterName.indexOf(':') > -1) {
			throw new ConfigurationException(
					"Invalid parameter name. Cannot contain colon.");
		}
		if (parameterName.indexOf('\'') > -1) {
			throw new ConfigurationException(
					"Invalid parameter name. Cannot contain apostrophe.");
		}
		if (parameterName.indexOf('?') > -1) {
			throw new ConfigurationException(
					"Invalid parameter name. Cannot contain question mark.");
		}
	
	}

	/**
	 * Count number of times a char appears in a string.
	 * @param string
	 * @param theChar
	 * @param end
	 * @return number of times char appear in string.
	 */
	private static int countChar(StringBuffer string, char theChar, int end) {
		int count = 0;
		for (int i = 0; i < end; i++) {
			if (string.charAt(i) == theChar) {
				count++;
			}
		}
		return count;
	}

	/**
	 * Find the smallest number in array between floor and ceiling, both
	 * inclusive. If none exist, ceiling will be returned.
	 * 
	 * @param array
	 * @param floor
	 * @param ceiling
	 * @return smallest number in array between floor and ceiling.
	 */
	private static int findLowest(int[] array, int floor, int ceiling) {
		int least = ceiling;
		for (int i = 0; i < array.length; i++) {
			if (array[i] < least && array[i] >= floor) {
				least = array[i];
			}
		}
		return least;
	}

	/**
	 * Parse the SQL statement containing host parameters. Note their position
	 * and replace with '?' for PreparedStatement. Return list with parameter names.
	 * @param statement
	 * @return Parameter list
	 * @throws ConfigurationException
	 */
	private static List findParameters(
			final StringBuffer statement) throws ConfigurationException {
	
		ArrayList parameterNames = new ArrayList();
		int[] possibleBreaks = new int[BREAK_CHARS.length];
		int str = 0;
		while ((str = statement.indexOf(":", str)) > -1) {
			if (isInsideQuotes(statement, str)) {
				str++;
				continue;
			}
			for (int i = 0; i < possibleBreaks.length; i++) {
				possibleBreaks[i] = statement.indexOf(BREAK_CHARS[i], str);
			}
	
			int end = findLowest(possibleBreaks, 0, statement.length());
			String name = statement.substring(str + 1, end);
	
			checkParameterName(name);
	
			parameterNames.add(name);
			statement.replace(str, end, "?");
		}
		return parameterNames;
	}

	
	/**
	 * Discover replacement properties in statement.
	 * String position of those values will be put into private 
	 * list. 
	 * @param statement
	 * @return Text replacement positions
	 */
	private static SortedMap findTextReplacements(StringBuffer statement) {
		TreeMap textReplacements = new TreeMap();
		int str = 0;
		while ((str = statement.indexOf("{{", str)) > -1) {
			int end = statement.indexOf("}}", str);
			String replacementKey = statement.substring(str + 2, end);
			statement.replace(str, end + 2, "");
			textReplacements.put(new Integer(str), replacementKey);
		}
		return textReplacements;
	}

	/**
	 * @param string
	 * @param end
	 * @return <code>true</code> if the end of string is inside quotes.
	 */
	private static boolean isInsideQuotes(StringBuffer string, int end) {
		return countChar(string, '\'', end) % 2 != 0;
	}
	
	private final List parameterNames;
	private final SortedMap replacementPositions;
	private final String sqlStatement;
	private final ImmutableSQL immutableSQL;
	
	SQLParser(StringBuffer staticSql) {
		this.parameterNames = findParameters(staticSql);
		this.replacementPositions = findTextReplacements(staticSql);
		this.sqlStatement = staticSql.toString();
		if (this.replacementPositions.isEmpty()) {
			this.immutableSQL = new ImmutableSQL(this.sqlStatement, this.parameterNames);
		} else {
			this.immutableSQL = null;
		}
	}

	/**
	 * Insert replacement strings.
	 * @param statement
	 * @param replacements
	 * @return Finished statement
	 */
	ImmutableSQL getRunnableStatement(TextReplacements replacements) {
		if (this.immutableSQL != null) {
			return this.immutableSQL;
		}
		
		StringBuffer sql = new StringBuffer(this.sqlStatement);
		Iterator properties = this.replacementPositions.entrySet().iterator();
		int valueLengths = 0;
		while (properties.hasNext()) {
			Map.Entry entry = (Map.Entry) properties.next();
			Integer position = (Integer) entry.getKey();
			String key = (String) entry.getValue();
			String value = replacements.getProperty(key);
			if (value == null) {
				String msg = "Text replacement '" + entry.getValue()
						+ "' has not been set.";
				throw new BrokerException(msg);
			}
			sql.insert(position.intValue() + valueLengths, value);
			valueLengths += value.length();
		}
		return new ImmutableSQL(sql.toString(), this.parameterNames);
	}

}
