package net.sourceforge.orbroker;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;

import net.kildenpedersen.reflect.Reflector;



/*
 * Created on Aug 1, 2004
 */

/**
 * @author Nils Kilden-Pedersen
 */
abstract class Statement {
    /**
     * Statement type enumeration.
     * @author Nils Kilden-Pedersen
     */
    static final class Type {
        private static final Map types = new HashMap();

        /**
         * <a href="http://freemarker.sourceforge.net/">FreeMarker</a> statement.
         */
        public static final Type FREEMARKER = new Type("FreeMarker");
        /**
         * Static SQL statement.
         */
        public static final Type STATIC = new Type("static");
        /**
         * <a href="http://jakarta.apache.org/velocity/">Velocity</a> statement.
         */
        public static final Type VELOCITY = new Type("Velocity");
        
        static {
            types.put(STATIC.toString(), STATIC);
            types.put(VELOCITY.toString(), VELOCITY);
            types.put(FREEMARKER.toString(), FREEMARKER);
        }
        
        private static boolean isFreeMarker(String statement) {
        	return (
        			(
        					statement.indexOf("<#if") > -1 &&
							statement.indexOf("</#if>") > -1
        			) ||
					(
        					statement.indexOf("<#list") > -1 &&
							statement.indexOf("</#list>") > -1
					)
        	);
        }
        
        private static boolean isVelocity(String statement) {
        	return (
        		(		statement.indexOf("#if") > -1 ||
        				statement.indexOf("#foreach") > -1
				) && statement.indexOf("#end") > -1
        	);
        }
        
        static Type evaluateStatement(String statement) {
        	if (isFreeMarker(statement)) {
        		return FREEMARKER;
        	} else if (isVelocity(statement)) {
        		return VELOCITY;
        	}
        	return STATIC;
        }
    	
        static Type getInstance(String name) throws NoSuchFieldException {
            if (name == null || name.trim().length() == 0) {
                return STATIC;
            }
            if (!types.containsKey(name)) {
                throw new NoSuchFieldException("Invalid statement type '" + name + "'");
            }
            return (Type) types.get(name);
        }
        
    	private final String typeName;
    	
    	private Type(String typeName) {
    		this.typeName = typeName;
    	}
        
    	/**
    	 * @inheritDoc
    	 * @see java.lang.Object#toString()
    	 */
    	public String toString() {
    		return this.typeName;
    	}
    }
	private static final char[] REPLACEMENT_CHARS = new char[] {'\t', '\f', '\r', '\n'};
    private static String cleanStatement(String statement) {
        statement = statement.trim();
        for (int i = 0; i < REPLACEMENT_CHARS.length; i++) {
            statement = statement.replace(REPLACEMENT_CHARS[i], ' ');
        }
        return statement;
    }

    private static ConfigurationException newUnknownTypeException(String id, Type type) {
    	return new ConfigurationException(
    	        "Code incomplete to handle statement '" + id + "' of type '" + type + "'");
    }

    /**
     * @param externalSource
     * @return source content
     * @throws IOException
     */
    private static String readExternalSource(final String externalSource) {
        BufferedReader reader = new BufferedReader(
                new InputStreamReader(
                        Statement.class.getResourceAsStream(externalSource)));
        StringBuffer buffer = new StringBuffer();
        String line = null;
        try {
            while ((line = reader.readLine()) != null) {
                buffer.append(line).append(' ');
            }
        } catch (IOException e) {
            throw new ConfigurationException(e);
        }
        return buffer.toString();
    }

	static Statement newInstance(
            String id,
            ResultObjectDefinition resultObjectDef,
            String externalSource) {
        
        String statement;
        try {
            statement = readExternalSource(externalSource);
        } catch (Exception e) {
            throw new ConfigurationException("Cannot read external source: " + externalSource, e);
        }
        return newInstance(id, statement, resultObjectDef);
    }

    static Statement newInstance(
    		String id, 
    		String statement, 
    		ResultObjectDefinition resultObjectDef) {

        statement = cleanStatement(statement);
        Statement.Type type = Statement.Type.evaluateStatement(statement);
        Statement instance;
    	if (Type.STATIC.equals(type)) {
    		if (StoredProcedureCall.isProcedureCall(statement)) {
    			instance = new StoredProcedureCall(id, statement, resultObjectDef);
    		} else {
    		    instance = new StaticStatement(id, statement, resultObjectDef);
    		}
    	} else if (Type.VELOCITY.equals(type)) {
    	    instance = new VelocityStatement(id, statement, resultObjectDef);
    	} else if (Type.FREEMARKER.equals(type)) {
    	    instance = new FreeMarkerStatement(id, statement, resultObjectDef);
    	} else {
    	    throw newUnknownTypeException(id, type);
    	}
    	return instance;
    }
    /**
     * Set row to map.
     * @param map Map to set values on
     * @param row Result row
     */
    static void setValuesOnMap(Map map, ResultRow row) {
        int[] readable = row.getReadableColumns();
        for (int i = 0; i < readable.length; i++) {
            int column = readable[i];
            String name = row.getColumnName(column);
            Object value = row.getObject(column);
            Reflector.setLeafValue(map, name, value);
        }
    }
    
    private final String id;
    private final ResultObjectDefinition resultObjectDef;
	/**
     * Constructor.
     * @param id Statement ID
     * @param resultObjectDef Result object factory
     */
    protected Statement(String id, ResultObjectDefinition resultObjectDef) {
		this.id = id;
		this.resultObjectDef = resultObjectDef;
	}

	/**
	 * @param context
	 * @param ps
	 * @param parsed
	 * @throws SQLException
	 */
	private void attachParameterValues(ConnectionContext context,
			PreparedStatement ps, ImmutableSQL parsed) throws SQLException {
	
		for (int index = 0, column = 1; 
				index < parsed.getParameterCount();
				index++, column++) {
			String treePath = parsed.getParameterName(index);
			Object parmValue = context.getParameterValue(treePath);
			setColumnValue(ps, parmValue, column);
		}
	}

	private PreparedStatement buildPreparedStatement(
			Connection con,
			boolean useScrollable,
			ImmutableSQL parsed) throws SQLException, BrokerException {
	
		String statement = parsed.getSqlStatement();
		logSqlStatement(statement);
		return con.prepareStatement(
				statement, 
				Broker.getResultSetType(useScrollable), 
				ResultSet.CONCUR_READ_ONLY);
	}
    
    /**
	 * Return the correct result-object builder.
	 * @param row
	 * @return resultObjectDef or <code>null</code> if none has been
	 *              assigned.
	 * @throws ConfigurationException
	 */
	private ResultObjectDefinition getRuntimeResultObjectDefinition(ResultRow row) throws ConfigurationException {

		if (this.resultObjectDef == null) {
			return null;
		}
		return this.resultObjectDef.getRuntimeResultObjectDef(row);
	}

	private QueryResult runStatement(
			Connection con,
			boolean useScrollable,
			ConnectionContext context, 
			ImmutableSQL runnableSQL) throws SQLException {
		
		java.sql.Statement stm = con.createStatement(
				Broker.getResultSetType(useScrollable), 
				ResultSet.CONCUR_READ_ONLY);
		String sql = runnableSQL.getSqlStatement();
		logSqlStatement(sql);
		ResultSet resultSet = stm.executeQuery(sql);
		return new QueryResult(stm, resultSet, context);
	}
	
    /**
     * Build result object. This will build a result object according to
     * the result object id specified on the statement. If no result
     * object id has been specified, then one of the two following objects
     * will be returned:
     * <ol>
     * <li>
     * 	If the ResultRow contains only one column,
     *		then that column value object will be returned. 
     * </li>
     * <li>
     * 	If the ResultRow contains more than one column,
     * 	then a {@link Map} will be returned containing the
     * 	column names as keys, and the column value objects
     * 	as values. 
     * </li>
     * </ol>
     * @param row Result row
     * @return result object
     */
    protected final Object buildResultObject(ResultRow row) {
    	
    	ResultObjectDefinition runtimeDef = getRuntimeResultObjectDefinition(row);
    	if (runtimeDef != null) {
    	    return runtimeDef.buildObject(row);    
    	}
    	
        if (row.getReadableColumns().length == 1) {
            int column = row.getReadableColumns()[0];
            return row.getObject(column);
        }
    	Map map = new HashMap();
    	setValuesOnMap(map, row);
    	return map;
    	
    }

    protected final void buildResultObject(ResultRow row, Object resultObject) 
    		throws ConfigurationException {
    
    	ResultObjectDefinition runtimeDef = getRuntimeResultObjectDefinition(row);
    	if (runtimeDef != null) {
    	    runtimeDef.setValues(resultObject, row);
    	    return;
    	}
    
    	if (resultObject instanceof Map) {
    		setValuesOnMap((Map) resultObject, row);
    		return;
    	}
    
    	String msg =
    		"No result-object has been assigned to " + toString();
    	throw new ConfigurationException(msg);
    
    }

    protected final void closeStatement(PreparedStatement ps) {
    	
        try {
            ps.close();
        } catch (SQLException e) {
        	Broker.log(Level.WARNING, e.toString());
        }
    }

    protected final void logSqlStatement(String statement) {
	    if (Broker.isLoggable(Level.INFO)) {
	        StringBuffer message = new StringBuffer();
	        message.append("Executing statement '").append(this.id).append("'...\n");
	        message.append(statement);
	        Broker.log(Level.INFO,  message.toString());
	    }
	}


    protected final void setColumnValue(
            PreparedStatement ps, 
            Object value,
			int columnIndex) throws SQLException {
	    ps.setObject(columnIndex, value);
	}
    /**
     * @param con
     * @param context
     * @param batchParameterName
     * @param batchParameters
     * @return rows updated per parameter
     * @throws SQLException
     */
    int[] executeBatch(
    		Connection con, 
    		ConnectionContext context,
    		String batchParameterName, 
    		Object[] batchParameters) throws SQLException {
    	
    	ImmutableSQL runtimeSQL = getRunnableSQL(context);
    	PreparedStatement ps = 
    		buildPreparedStatement(con, false, runtimeSQL);

        try {
        	for (int i = 0; i < batchParameters.length; i++) {
        	    context.setParameter(batchParameterName, batchParameters[i]);
        		attachParameterValues(context, ps, runtimeSQL);
        		ps.addBatch();
        	}
        	context.removeParameter(batchParameterName);
        	return ps.executeBatch();
        } finally {
            closeStatement(ps);
        }
    }

    /**
     * @param con
     * @param scrollableSupport
     * @param context
     * @inheritDoc
     * @return query result 
     * @throws SQLException
     */
    QueryResult executeQuery(
    		Connection con, 
    		boolean scrollableSupport,
    		ConnectionContext context) throws SQLException {
    	
    	ImmutableSQL runtimeSQL = getRunnableSQL(context);
    	
    	if (!runtimeSQL.hasParameters()) {
    		return runStatement(con, scrollableSupport, context, runtimeSQL);
    	}

    	PreparedStatement ps = 
    		buildPreparedStatement(con, scrollableSupport, runtimeSQL);
    	attachParameterValues(context, ps, runtimeSQL);
    	ResultSet rs = ps.executeQuery();
    	return new QueryResult(ps, rs, context);
    }

    /**
     * @param con
     * @param context
     * @inheritDoc
     * @return number of records affected
     * @throws SQLException
     */
    int executeUpdate(
    		Connection con, 
    		ConnectionContext context)
    		throws SQLException {
    
    	ImmutableSQL runtimeSQL = getRunnableSQL(context);
    	PreparedStatement ps = 
    		buildPreparedStatement(con, false, runtimeSQL);
        try {
        	attachParameterValues(context, ps, runtimeSQL);
        	return ps.executeUpdate();
        } finally {
            closeStatement(ps);
        }
    }
    
    /**
     * @return Returns id.
     */
    final String getId() {
        return this.id;
    }

	abstract ImmutableSQL getRunnableSQL(ConnectionContext context) throws BrokerException;
	
	/**
	 * @inheritDoc
	 * @see java.lang.Object#toString()
	 */
	public final String toString() {
		return "Statement: " + this.id;
	}
}
