/*
 * Created on Jun 24, 2004
 */
package net.sourceforge.orbroker;

import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.ParameterMetaData;
import java.sql.ResultSet;
import java.sql.SQLException;

/**
 * @author Nils Kilden-Pedersen
 */
final class StoredProcedureCall extends StaticStatement {

    /**
     * Is statement a stored procedure call?
     * @param statement
     * @return true, if it's a CALL
     */
    static boolean isProcedureCall(String statement) {
        if (statement.length() > 4) {
            statement = statement.substring(0, 4).toUpperCase();
        }
        return 
        	statement.startsWith("CALL") ||
        	statement.startsWith("{") ||
        	statement.startsWith(":") ||
			statement.startsWith("}");
    }

    /**
	 * Constructor.
	 * @param id
	 * @param statement
	 * @param resultObjectDef
	 */
	StoredProcedureCall(String id, String statement,
			ResultObjectDefinition resultObjectDef) {
		super(id, statement, resultObjectDef);
	}

	/**
	 * For stored procedure calls.
	 * @param context
	 * @param cs
	 * @param parsed
	 * @return Parameter OUTPUT status
	 * @throws SQLException
	 */
	private boolean[] attachParameterValues(
	        ConnectionContext context,
			CallableStatement cs,
			ImmutableSQL parsed) throws SQLException {

		ParameterMetaData metaData = cs.getParameterMetaData();
		int parmCount = metaData.getParameterCount();
		if (parmCount != parsed.getParameterCount()) {
			throw new BrokerException(
					"Metadata parameter count of " + 
					parmCount + " does not match actual count of " + parsed.getParameterCount());
		}

		boolean[] isParameterOutput = new boolean[parmCount];
		for (int i = 0, parameterIndex = 1; i < parmCount; i++, parameterIndex++) {
			String treePath = parsed.getParameterName(i);
			int mode = metaData.getParameterMode(parameterIndex);
			if (mode == ParameterMetaData.parameterModeIn
					|| mode == ParameterMetaData.parameterModeInOut) {
				Object parmValue = context.getParameterValue(treePath);
				setColumnValue(cs, parmValue, parameterIndex);
			}
			isParameterOutput[i] = (
			        mode == ParameterMetaData.parameterModeInOut ||
			        mode == ParameterMetaData.parameterModeOut
			);
			if (isParameterOutput[i]) {
			    cs.registerOutParameter(
			            parameterIndex, metaData.getParameterType(parameterIndex));
			}
		}
		return isParameterOutput;
	}

    private CallableStatement buildCallableStatement (
			Connection con,
			boolean scrollableSupport,
			ImmutableSQL parsedSQL) throws SQLException {

		String statement = parsedSQL.getSqlStatement();
		logSqlStatement(statement);
		return con.prepareCall(
				statement, 
				Broker.getResultSetType(scrollableSupport),
				ResultSet.CONCUR_READ_ONLY);
	}

    /**
     * @param context
     * @param cs
     * @param parsedSQL
     * @param isParameterOutput
     * @throws SQLException
     */
    private void setOutputParameters(
    		ConnectionContext context, 
			CallableStatement cs,
			ImmutableSQL parsedSQL,
			boolean[] isParameterOutput) throws SQLException {
        ResultSet resultSet = 
        	new ResultParameters(
        			cs, isParameterOutput, 
					parsedSQL.copyParameterNames());
		ResultRow row = new ResultRow(resultSet, context, cs.getConnection());
		setValuesOnMap(context.getParametersAsMap(), row);
    }

    /**
     * @inheritDoc
     * @throws SQLException
     */
    int[] executeBatch(
    		Connection con, 
    		ConnectionContext context,
    		String batchParameterName, 
    		Object[] batchParameters) throws SQLException {
    	
    	ImmutableSQL sql = getRunnableSQL(context);
    	
    	CallableStatement cs = buildCallableStatement(con, false, sql);
    	try {
        	for (int i = 0; i < batchParameters.length; i++) {
        	    context.setParameter(batchParameterName, batchParameters[i]);
        		attachParameterValues(context, cs, sql);
        		cs.addBatch();
        	}
        	context.removeParameter(batchParameterName);
        	return cs.executeBatch();
        } finally {
            closeStatement(cs);
        }
    }

    /**
	 * 
	 * @inheritDoc
	 * @throws SQLException
	 */
	QueryResult executeQuery(
			Connection con, 
			boolean scrollableSupport,
			ConnectionContext context) throws SQLException {
		
		ImmutableSQL parsedSQL = getRunnableSQL(context);
		CallableStatement cs = 
			buildCallableStatement(con, scrollableSupport, parsedSQL);
		boolean[] isParameterOutput = attachParameterValues(context, cs, parsedSQL);
		ResultSet resultSet;
		cs.execute();
		
		try {
			resultSet = cs.getResultSet();
		} catch (SQLException e) {
			resultSet = null;
		}

        ResultSet outputParameters = 
            new ResultParameters(
            		cs, 
            		isParameterOutput, 
					parsedSQL.copyParameterNames()); 

        if (resultSet == null) {
            resultSet = outputParameters;
            outputParameters = null;
        }
        return new QueryResult(cs, resultSet, context, outputParameters);
	}

	/**
     * @inheritDoc
     * @see net.sourceforge.orbroker.Statement#executeUpdate(java.sql.Connection, net.sourceforge.orbroker.ConnectionContext)
     */
    int executeUpdate(Connection con, ConnectionContext context)
            throws SQLException {
		
    	ImmutableSQL parsedSql = getRunnableSQL(context);
        CallableStatement cs = buildCallableStatement(con, false, parsedSql);
        try {
    		boolean[] isParameterOutput = attachParameterValues(context, cs, parsedSql);
    		int rowsAffected = cs.executeUpdate();
    		setOutputParameters(context, cs, parsedSql, isParameterOutput);
    		return rowsAffected;
        } finally {
            closeStatement(cs);
        }
    }
}
