package net.sourceforge.orbroker;

import java.lang.reflect.Array;
import java.sql.Connection;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;



/*
 * Created on Mar 24, 2004
 *  
 */

/**
 * @author Nils Kilden-Pedersen
 */
final class SubQuery extends ValueType {

	private String statementID;
	private Broker broker;
	private Map valueTypes = new HashMap();
	private boolean isLazyCollection;

	SubQuery(Class type, String statementID, Broker broker, boolean lazy) {
		super(type);
		this.statementID = statementID;
		this.broker = broker;
		this.isLazyCollection = lazy;
	}

	void addParameterValue(String name, Object value) {
		this.valueTypes.put(name, value);
	}
	/**
	 * @see net.sourceforge.orbroker.ValueType#getValue(net.sourceforge.orbroker.ResultRow)
	 */
	Object getValue(ResultRow row) {
		ConnectionContext context = new ConnectionContext();
		context.setTextReplacements(row.getTextReplacements());
		context.setParameters(getParameterValues(row));

		if (!this.isLazyCollection) {
			return getImmediateValue(context, row.getConnection());
		}
		
		Class thisType = getType();
		if (SortedSet.class.equals(thisType)) {
		    return new LazySortedSet(this.broker, this.statementID, context);
		} else if (Set.class.equals(thisType)) {
			return new LazySet(this.broker, this.statementID, context);
		} else if (List.class.equals(thisType) || Collection.class.equals(thisType)) {
			return new LazyList(this.broker, this.statementID, context);
		} 
		String msg = "Lazy collections only works on Collection interfaces";  
		throw new ConfigurationException(msg);
	}
	
	private Parameters getParameterValues(ResultRow row) {
		Parameters values = new Parameters();
		Iterator varEntries = this.valueTypes.entrySet().iterator();
		while (varEntries.hasNext()) {
			Map.Entry entry = (Map.Entry) varEntries.next();
			String varName = (String) entry.getKey();
			ValueType varType = (ValueType) entry.getValue();
			Object value = varType.getValue(row);
			values.addParameter(varName, value);
		}
		return values;
	}

	private Object getImmediateValue(ConnectionContext context, Connection con) {
		
		Object result;
		Executable exe = this.broker.obtainExecutable(con);
		exe.addContext(context);
		Class requiredType = getType();
		Class nonPrimitiveArrayType = null;
		if (requiredType.isArray()) {
			nonPrimitiveArrayType = requiredType.getComponentType();
			if (nonPrimitiveArrayType.isPrimitive()) {
				nonPrimitiveArrayType = null;
			}
		}
		try {
			if (Collection.class.isAssignableFrom(requiredType)) {
				result = createCollectionInstance();
				exe.selectMany(this.statementID, (Collection) result);
			} else if (nonPrimitiveArrayType != null) {
				List list = exe.selectMany(this.statementID);
				result = Array.newInstance(nonPrimitiveArrayType, list.size());
				list.toArray((Object[]) result);
			} else {
				result = exe.selectOne(this.statementID);
			}
		} finally {
			/*	Do not use Broker.releaseExecutable
			 	since it clears cache for connection. */
			exe.release();
		}
		return result;
	}
	
	private Collection createCollectionInstance() {
		
		if (getType().isInterface()) {
			if (Collection.class.equals(getType()) ||
				List.class.isAssignableFrom(getType())) {
				return new ArrayList();
			} else if (Set.class.isAssignableFrom(getType())) {
				return new HashSet();
			} else if (SortedSet.class.isAssignableFrom(getType())) {
				return new TreeSet();
			}
		}
		
		try {
			return (Collection) getType().newInstance();
		} catch (InstantiationException e) {
			// Throw exception below
		} catch (ClassCastException e) {
			assert false : "Method should only be called after checking type is Collection.";
		} catch (Exception e) {
			throw new ReflectionException(e);
		}
		String msg = "Cannot create " + getType();
		throw new ConfigurationException(msg);
	}

}
