/*
 * Created on Feb 22, 2004
 */
package net.sourceforge.orbroker;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.logging.Level;



/**
 * Any database modifications (INSERT, UPDATE, DELETE) must be done through
 * the Transaction object. It wraps a {@link java.sql.Connection Connection}
 * object, and like a {@link Query} instance, {@link #close()} must be called after
 * use (in a <code>finally</code> block), to avoid connection pool leaks.
 * A transaction using an isolation level of 
 * {@link java.sql.Connection#TRANSACTION_NONE}
 * is considered non-transactional and will have auto-commit enabled. This makes
 * it illegal to call either {@link #commit()} or {@link #rollback()}. All other isolation
 * levels require a call to {@link #commit()} method for the transaction to commit.
 * 
 * @author Nils Kilden-Pedersen
 * @see net.sourceforge.orbroker.Broker#startTransaction()
 * @see net.sourceforge.orbroker.Broker#startTransaction(int)
 */
public final class Transaction extends ExecutableConnection {

	private static void log(String message) {
	    Broker.log(Level.INFO, message);
	}
	
    private ArrayList executables;
	private boolean transactionStarted = false;
	Transaction(Broker broker, Integer isolationLevel) throws BrokerException {
		super(broker, isolationLevel);
	}

	/**
	 * 
	 */
	private void releaseExecutables() {
		if (this.executables != null) {
    		int size = this.executables.size();
    		for (int i = 0; i < size; i++) {
				Executable exe = (Executable) this.executables.get(i);
				exe.release();
			}
    	}
	}
	/**
	 * @see net.sourceforge.orbroker.QueryableConnection#getClosedConnectionMessage()
	 */
	protected String getClosedConnectionMessage() {
		return "Transaction has been closed.";
	}

	/**
	 * @see net.sourceforge.orbroker.QueryableConnection#getFinalizeWarning()
	 */
	protected String getFinalizeWarning() {
		return "A Transaction was not closed. Temporary connection pool leak now closed.";
	}

	/**
	 * @see net.sourceforge.orbroker.QueryableConnection#setConnectionProperties()
	 */
	protected void setConnectionProperties() throws SQLException {
		Connection con = getActiveConnection();
		if (con.isReadOnly()) {
			con.setReadOnly(false);
		}
	}

    /**
     * @see net.sourceforge.orbroker.BrokerConnection#shouldConnectionClose()
     */
    protected boolean shouldConnectionClose() {
        return true;
    }

    /**
     * Close transaction. If Transaction has not been explicity committed
     * or rolled back, it will automatically be rolled back on close. This can
     * simplify code by automatically rolling back on close:
     * <pre>Transaction txn = broker.startTransaction();
     * try {
     *    txn.setParameter("account", account);
     *    txn.execute("updateAccount");
     *    txn.commit();
     * ] finally {
     *    txn.close();
     * }</pre>
     * There is no need to issue a rollback. If commit() is not called due to an exception,
     * rollback is done automatically.
     * This is assuming that the transaction is in isolation level other than
     * {@link Connection#TRANSACTION_NONE}.
     */
    public void close() {
    	if (this.transactionStarted) {
    		rollback();
    	}
    	releaseExecutables();
    	dropConnection();
    }

	/**
	 * Commit the transaction. 
	 * <p>This method cannot be called if isolation level
	 * is {@link Connection#TRANSACTION_NONE}.</p>
	 * @throws BrokerException
	 * @throws ConstraintException
	 * @throws DeadlockException
	 * @see #rollback()
	 */
	public void commit() throws BrokerException, ConstraintException, DeadlockException {
		if (isAutoCommit()) {
			throw new BrokerException(
					"Commit not supported with transaction isolation level of " + 
					getTransactionIsolationText()
					);
		}

		/* Since transaction is automatically rolled back
		 * on a failed commit, it is safe to set 
		 * this.transactionStarted to false before trying.
		 */
		this.transactionStarted = false;
		try {
			getActiveConnection().commit();
		} catch (SQLException e) {
			throwException(e);
		}
		log("Transaction committed.");
	}

	/**
	 * Obtain an Executable that uses this Transaction's connection. 
	 * This class provides safe semantics for other methods that 
	 * need to take part in the transaction, without exposing commit 
	 * or rollback capabilities.
	 * The Executable will be automatically released and invalidated when 
	 * Transaction is closed.
     *
	 * @return Executable
	 * @see #close()
	 * @see Broker#obtainExecutable(Connection)
	 */
	public Executable obtainExecutable() {
		if (this.executables == null) {
			this.executables = new ArrayList(5);
		}
		Executable executable = newExecutable(this);
		this.executables.add(executable);
	    return executable;
	}
	
	/**
	 * Roll back the transaction. If rollback() is not called on an
	 * exception, it will be done automatically on {@link #close()} 
	 * (provided that {@link #close()} method is called in a 
	 * <code>finally</code> block as it always should).
	 * <p>This method cannot be called if isolation level
	 * is {@link Connection#TRANSACTION_NONE}.</p>
	 * @throws BrokerException
	 * @see #commit()
	 */
	public void rollback() throws BrokerException {
		if (isAutoCommit()) {
			throw new BrokerException(
					"Rollback not supported with transaction isolation level of " + 
					getTransactionIsolationText()
					);
		}

		/* If a rollback fails, I'm not really sure what
		 * state the transaction is in. If it's committed then
		 * it is safe to set this.transactionStarted to false,
		 * if not, the intent is to rollback and that cannot
		 * happen, so the transaction must be considered
		 * finished.
		 */
		this.transactionStarted = false;
		try {
			getActiveConnection().rollback();
		} catch (SQLException e) {
		    throwException(e);
		}
		log("Transaction rolled back.");
	}

	/**
	 * @inheritDoc
	 * @see net.sourceforge.orbroker.ExecutableConnection#markTransactionStarted()
	 */
	protected void markTransactionStarted() {
		if (!this.transactionStarted && !isAutoCommit()) {
			this.transactionStarted = true;
			log("Transaction started...");
		}
	}
	
	
}